"""Linear Operator Matrix class."""

from __future__ import annotations

import operator
from collections.abc import Callable, Iterator, Sequence
from functools import cached_property, reduce
from types import EllipsisType
from typing import cast

import torch
from typing_extensions import Self, Unpack

from mrpro.operators.LinearOperator import LinearOperator, LinearOperatorSum
from mrpro.operators.Operator import Operator
from mrpro.operators.ZeroOp import ZeroOp

_SingleIdxType = int | slice | EllipsisType | Sequence[int]
_IdxType = _SingleIdxType | tuple[_SingleIdxType, _SingleIdxType]


class LinearOperatorMatrix(Operator[Unpack[tuple[torch.Tensor, ...]], tuple[torch.Tensor, ...]]):
    r"""Matrix of linear operators.

    A matrix of linear operators, where each element is a `~mrpro.operators.LinearOperator`.

    This matrix can be applied to a sequence of tensors, where the number of tensors should match
    the number of columns of the matrix. The output will be a sequence of tensors, where the number
    of tensors will match the number of rows of the matrix.
    The i-th output tensor is calculated as
    :math:`\sum_j \text{operators}[i][j](x[j])` where :math:`\text{operators}[i][j]` is the linear operator
    in the i-th row and j-th column and :math:`x[j]` is the j-th input tensor.

    The matrix can be indexed and sliced like a regular matrix to get submatrices.
    If indexing returns a single element, it is returned as a `~mrpro.operators.LinearOperator`.

    Basic arithmetic operations are supported with `~mrpro.operators.LinearOperator` and Tensors.
    Use ``A | B`` for horizontal stacking (side-by-side columns) and ``A % B`` for vertical stacking (stacked rows).

    """

    _operators: list[list[LinearOperator]]

    def __init__(self, operators: Sequence[Sequence[LinearOperator]]):
        """Initialize Linear Operator Matrix.

        Create a matrix of `~mrpro.operators.LinearOperator` from a sequence of rows, where each row is a sequence
        of `~mrpro.operators.LinearOperator` that represent the columns of the matrix.

        Parameters
        ----------
        operators
            A sequence of rows, which are sequences of `~mrpro.operators.LinearOperator`.
        """
        if not all(isinstance(op, LinearOperator) for row in operators for op in row):
            raise ValueError('All elements should be LinearOperators.')
        if not all(len(row) == len(operators[0]) for row in operators):
            raise ValueError('All rows should have the same length.')
        super().__init__()
        self._operators = cast(  # cast because ModuleList is not recognized as a list
            list[list[LinearOperator]], torch.nn.ModuleList(torch.nn.ModuleList(row) for row in operators)
        )
        self._shape = (len(operators), len(operators[0]) if operators else 0)

    @property
    def shape(self) -> tuple[int, int]:
        """Shape of the Operator Matrix (rows, columns)."""
        return self._shape

    def __call__(self, *x: torch.Tensor) -> tuple[torch.Tensor, ...]:
        """Apply the linear operator matrix to a sequence of input tensors.

        The i-th output tensor is calculated as the sum over j of
        `operators[i][j](x[j])`, where `operators[i][j]` is the
        linear operator in the i-th row and j-th column, and `x[j]` is
        the j-th input tensor.

        Parameters
        ----------
        *x
            Input tensors. The number of input tensors must match the
            number of columns in the operator matrix.

        Returns
        -------
            Output tensors. The number of output tensors will match the
            number of rows in the operator matrix.
        """
        return super().__call__(*x)

    def forward(self, *x: torch.Tensor) -> tuple[torch.Tensor, ...]:
        """Apply forward of LinearOperatorMatrix.

        .. note::
            Prefer calling the instance of the LinearOperatorMatrix operator as ``operator(x)`` over
            directly calling this method. See this PyTorch `discussion <https://discuss.pytorch.org/t/is-model-forward-x-the-same-as-model-call-x/33460/3>`_.
        """
        if len(x) != self.shape[1]:
            raise ValueError('Input should be the same number of tensors as the LinearOperatorMatrix has columns.')
        return tuple(
            reduce(operator.add, (op(xi)[0] for op, xi in zip(row, x, strict=True))) for row in self._operators
        )

    def __getitem__(self, idx: _IdxType) -> Self | LinearOperator:
        """Index the Operator Matrix.

        Parameters
        ----------
        idx
            Index or slice to select rows and columns.

        Returns
        -------
            Subset `LinearOperatorMatrix` or `~mrpro.operators.LinearOperator`.
        """
        idxs: tuple[_SingleIdxType, _SingleIdxType] = idx if isinstance(idx, tuple) else (idx, slice(None))
        if len(idxs) > 2:
            raise IndexError('Too many indices for LinearOperatorMatrix')

        def _to_numeric_index(idx: slice | int | Sequence[int] | EllipsisType, length: int) -> Sequence[int]:
            """Convert index to a sequence of integers or raise an error."""
            if isinstance(idx, slice):
                if (idx.start is not None and (idx.start < -length or idx.start >= length)) or (
                    idx.stop is not None and (idx.stop < -length or idx.stop > length)
                ):
                    raise IndexError('Index out of range')
                return range(*idx.indices(length))
            if isinstance(idx, int):
                if idx < -length or idx >= length:
                    raise IndexError('Index out of range')
                return (idx,)
            if idx is Ellipsis:
                return range(length)
            if isinstance(idx, Sequence):
                if min(idx) < -length or max(idx) >= length:
                    raise IndexError('Index out of range')
                return idx
            else:
                raise IndexError('Invalid index type')

        row_numbers = _to_numeric_index(idxs[0], self._shape[0])
        col_numbers = _to_numeric_index(idxs[1], self._shape[1])

        sliced_operators = [
            [row[col_number] for col_number in col_numbers]
            for row in [self._operators[row_number] for row_number in row_numbers]
        ]

        # Return a single operator if only one row and column is selected
        if len(row_numbers) == 1 and len(col_numbers) == 1:
            return sliced_operators[0][0]
        else:
            return self.__class__(sliced_operators)

    def __iter__(self) -> Iterator[Sequence[LinearOperator]]:
        """Iterate over the rows of the Operator Matrix."""
        return iter(self._operators)

    def __repr__(self):
        """Representation of the Operator Matrix."""
        return f'LinearOperatorMatrix(shape={self._shape}, operators={self._operators})'

    # Note: The type ignores are needed because we currently cannot do arithmetic operations with non-linear operators.
    def __add__(self, other: Self | LinearOperator | torch.Tensor | complex) -> Self:  # type: ignore[override]
        """Addition."""
        operators: list[list[LinearOperator]] = []
        if isinstance(other, LinearOperatorMatrix):
            if self.shape != other.shape:
                raise ValueError('OperatorMatrix shapes do not match.')
            for self_row, other_row in zip(self._operators, other._operators, strict=False):
                operators.append([s + o for s, o in zip(self_row, other_row, strict=False)])
        elif isinstance(other, LinearOperator | torch.Tensor | complex):
            if not self.shape[0] == self.shape[1]:
                raise NotImplementedError('Cannot add a LinearOperator to a non-square OperatorMatrix.')
            for i, self_row in enumerate(self._operators):
                operators.append([op + other if i == j else op for j, op in enumerate(self_row)])
        else:
            return NotImplemented
        return self.__class__(operators)

    def __radd__(self, other: Self | LinearOperator | torch.Tensor | complex) -> Self:
        """Right addition."""
        return self.__add__(other)

    def __mul__(self, other: torch.Tensor | Sequence[torch.Tensor | complex] | complex) -> Self:
        """LinearOperatorMatrix*Tensor multiplication.

        Example: :math:`([A,B]c)(x) = [Ac, Bc](x) = A(cx) + B(cx)`
        """
        if isinstance(other, torch.Tensor | complex | float | int):
            other_: Sequence[torch.Tensor | complex] = (other,) * self.shape[1]
        elif len(other) != self.shape[1]:
            raise ValueError('Other should have the same length as the operator has columns.')
        else:
            other_ = other
        operators = []
        for row in self._operators:
            operators.append([op * o for op, o in zip(row, other_, strict=True)])
        return self.__class__(operators)

    def __rmul__(self, other: torch.Tensor | Sequence[torch.Tensor] | complex) -> Self:
        """Tensor*LinearOperatorMatrix multiplication.

        Example: (c[A,B])(x) = [cA, cB](x) = cA(x) + cB(x)
        """
        if isinstance(other, torch.Tensor | complex | float | int):
            other_: Sequence[torch.Tensor | complex] = (other,) * self.shape[0]
        elif len(other) != self.shape[0]:
            raise ValueError('Other should have the same length as the operator has rows.')
        else:
            other_ = other
        operators = []
        for row, o in zip(self._operators, other_, strict=True):
            operators.append([cast(LinearOperator, o * op) for op in row])
        return self.__class__(operators)

    def __rmatmul__(self, other: LinearOperator) -> Self:  # type: ignore[misc]
        """Composition of operators."""
        operators = [[op @ other for op in row] for row in self._operators]
        return self.__class__(operators)

    def __matmul__(self, other: LinearOperator | Self) -> Self:  # type: ignore[override]
        """Composition of operators."""
        if isinstance(other, LinearOperator):
            operators = [[other @ op for op in row] for row in self._operators]
            return self.__class__(operators)
        elif isinstance(other, LinearOperatorMatrix):
            if self.shape[1] != other.shape[0]:
                raise ValueError('OperatorMatrix shapes do not match.')
            new_operators = []
            for row in self._operators:
                new_row = []
                for other_col in zip(*other._operators, strict=True):
                    elements = [s @ o for s, o in zip(row, other_col, strict=True)]
                    new_row.append(LinearOperatorSum(*elements))
                new_operators.append(new_row)
            return self.__class__(new_operators)
        return NotImplemented

    @property
    def H(self) -> Self:  # noqa N802
        """Adjoints of the operators."""
        return self.__class__([[op.H for op in row] for row in zip(*self._operators, strict=True)])

    @cached_property
    def gram(self) -> Self:
        """Gram matrix of the operators."""
        n, m = self.shape
        operators: list[list[LinearOperator]] = [[ZeroOp() for _ in range(m)] for _ in range(m)]
        for j in range(m):
            operators[j][j] = reduce(operator.add, (self._operators[i][j].gram for i in range(n)), ZeroOp())
            for k in range(j + 1, m):
                operators[j][k] = reduce(
                    operator.add, (self._operators[i][j].H @ self._operators[i][k] for i in range(n)), ZeroOp()
                )
                operators[k][j] = operators[j][k].H
        return self.__class__(operators)

    def adjoint(self, *x: torch.Tensor) -> tuple[torch.Tensor, ...]:
        """Apply the adjoint of the linear operator matrix.

        This is achieved by applying the matrix composed of the adjoints of each
        individual operator (i.e., `self.H`) to the input tensors `*x`.
        The k-th output tensor is `sum_i self.operators[i][k].H(x[i])`.

        Parameters
        ----------
        *x
            Input tensors. The number of input tensors must match the
            number of rows in the original operator matrix.

        Returns
        -------
            Output tensors. The number of output tensors will match the
            number of columns in the original operator matrix.
        """
        return self.H(*x)

    @classmethod
    def from_diagonal(cls, *operators: LinearOperator) -> Self:
        """Create a diagonal `LinearOperatorMatrix`.

        Construct a square `LinearOperatorMatrix` with the given linear operators on the diagonal,
        resulting in a block-diagonal linear operator.

        Parameters
        ----------
        operators
            Sequence of `~mrpro.operators.LinearOperator` to be placed on the diagonal.
        """
        operator_matrix: list[list[LinearOperator]] = [
            [op if i == j else ZeroOp(False) for j in range(len(operators))] for i, op in enumerate(operators)
        ]
        return cls(operator_matrix)

    def operator_norm(
        self,
        *initial_value: torch.Tensor,
        dim: Sequence[int] | None = None,
        max_iterations: int = 20,
        relative_tolerance: float = 1e-4,
        absolute_tolerance: float = 1e-5,
        callback: Callable[[torch.Tensor], None] | None = None,
    ) -> torch.Tensor:
        r"""Upper bound of operator norm of the Matrix.

        Uses the bounds :math:`||[A, B]^T||\leq\sqrt{(||A||^2 + ||B||^2)}` and :math:`||[A, B]||\leq\max(||A||,||B||)`
        to estimate the operator norm of the matrix.
        First,  operator_norm is called on each element of the matrix.
        Next, the norm is estimated for each column using the first bound.
        Finally, the norm of the full matrix of linear operators is calculated using the second bound.

        Parameters
        ----------
        initial_value
            Initial value(s) for the power iteration, length should match the number of columns
            of the operator matrix.
        dim
            Dimensions to calculate the operator norm over. Other dimensions are assumed to be
            batch dimensions. None means all dimensions.
        max_iterations
            Maximum number of iterations used in the power iteration.
        relative_tolerance
            Relative tolerance for convergence.
        absolute_tolerance
            Absolute tolerance for convergence.
        callback
            Callback function to be called with the current estimate of the operator norm.


        Returns
        -------
        Estimated operator norm upper bound.
        """

        def _singlenorm(op: LinearOperator, initial_value: torch.Tensor):
            return op.operator_norm(
                initial_value,
                dim=dim,
                max_iterations=max_iterations,
                relative_tolerance=relative_tolerance,
                absolute_tolerance=absolute_tolerance,
                callback=callback,
            )

        if len(initial_value) != self.shape[1]:
            raise ValueError('Initial value should have the same length as the operator has columns.')
        norms = torch.stack(
            [
                torch.stack([_singlenorm(op, iv) for op, iv in zip(row, initial_value, strict=True)])
                for row in self._operators
            ]
        )
        norm = norms.square().sum(0).sqrt().amax(0)
        return norm

    def __or__(self, other: LinearOperator | LinearOperatorMatrix) -> Self:
        """Horizontal stacking."""
        if isinstance(other, LinearOperator):
            if (rows := self.shape[0]) > 1:
                raise ValueError(
                    f'Shape mismatch in horizontal stacking : cannot stack LinearOperator and matrix with {rows} rows.'
                )
            operators = [[*self._operators[0], other]]
            return self.__class__(operators)
        else:
            if (rows_self := self.shape[0]) != (rows_other := other.shape[0]):
                raise ValueError(
                    f'Shape mismatch in horizontal stacking: cannot stack matrices with {rows_self} and {rows_other}.'
                )
            operators = [[*self_row, *other_row] for self_row, other_row in zip(self, other, strict=True)]
            return self.__class__(operators)

    def __ror__(self, other: LinearOperator) -> Self:
        """Horizontal stacking."""
        if (rows := self.shape[0]) > 1:
            raise ValueError(
                f'Shape mismatch in horizontal stacking: cannot stack LinearOperator and matrix with {rows} rows.'
            )
        operators = [[other, *self._operators[0]]]
        return self.__class__(operators)

    def __mod__(self, other: LinearOperator | LinearOperatorMatrix) -> Self:
        """Vertical stacking."""
        if isinstance(other, LinearOperator):
            if (cols := self.shape[1]) > 1:
                raise ValueError(
                    f'Shape mismatch in vertical stacking:cannot stack LinearOperator and matrix with {cols} columns.'
                )
            operators = [*self._operators, [other]]
            return self.__class__(operators)
        else:
            if (cols_self := self.shape[1]) != (cols_other := other.shape[1]):
                raise ValueError(
                    'Shape mismatch in vertical stacking:'
                    f'cannot stack matrices with {cols_self} and {cols_other} columns.'
                )
            operators = [*self._operators, *other._operators]
            return self.__class__(operators)

    def __rmod__(self, other: LinearOperator) -> Self:
        """Vertical stacking."""
        if (cols := self.shape[1]) > 1:
            raise ValueError(
                f'Shape mismatch in vertical stacking: cannot stack LinearOperator and matrix with {cols} columns.'
            )
        operators = [[other], *self._operators]
        return self.__class__(operators)



"""2D radial trajectory class."""

import torch

from mrpro.data.KTrajectory import KTrajectory
from mrpro.data.traj_calculators.KTrajectoryCalculator import KTrajectoryCalculator
from mrpro.utils.reshape import unsqueeze_tensors_left


class KTrajectoryRadial2D(KTrajectoryCalculator):
    """Radial 2D trajectory."""

    def __init__(self, angle: float = torch.pi * 0.618034) -> None:
        """Initialize KTrajectoryRadial2D.

        Parameters
        ----------
        angle
            angle in rad between two radial lines
        """
        super().__init__()
        self.angle: float = angle

    def __call__(
        self,
        *,
        n_k0: int,
        k0_center: int | torch.Tensor,
        k1_idx: torch.Tensor,
        reversed_readout_mask: torch.Tensor | None = None,
        **_,
    ) -> KTrajectory:
        """Calculate radial 2D trajectory for given header information.

        Parameters
        ----------
        n_k0
            number of samples in k0 (radial readout)
        k0_center
            position of k-space center in k0
        k1_idx
            indices of k1 (angle)
        reversed_readout_mask
            boolean tensor indicating reversed readout

        Returns
        -------
            radial 2D trajectory for given KHeader
        """
        radial = self._readout(n_k0=n_k0, k0_center=k0_center, reversed_readout_mask=reversed_readout_mask)
        angle = k1_idx * self.angle
        kx = radial * torch.cos(angle)
        ky = radial * torch.sin(angle)
        kz = torch.zeros(1)
        kz, ky, kx = unsqueeze_tensors_left(kz, ky, kx, ndim=5)
        return KTrajectory(kz, ky, kx)



"""K-space trajectory from .seq file class."""

from pathlib import Path

import torch
from einops import repeat

from mrpro.data.KTrajectory import KTrajectory
from mrpro.data.SpatialDimension import SpatialDimension
from mrpro.data.traj_calculators.KTrajectoryCalculator import KTrajectoryCalculator


class KTrajectoryPulseq(KTrajectoryCalculator):
    """Trajectory from .seq file."""

    def __init__(self, seq_path: str | Path, repeat_detection_tolerance: None | float = 1e-3) -> None:
        """Initialize KTrajectoryPulseq.

        Parameters
        ----------
        seq_path
            absolute path to .seq file
        repeat_detection_tolerance
            tolerance for repeat detection when creating KTrajectory
        """
        super().__init__()
        self.seq_path = seq_path
        self.repeat_detection_tolerance = repeat_detection_tolerance

    def __call__(
        self,
        *,
        n_k0: int,
        encoding_matrix: SpatialDimension,
        **_,
    ) -> KTrajectory:
        """Calculate trajectory from given .seq file and header information.

        Parameters
        ----------
        n_k0
            number of samples in k0
        encoding_matrix
            encoding matrix

        Returns
        -------
            trajectory of type KTrajectory
        """
        from pypulseq import Sequence

        # calculate k-space trajectory using PyPulseq
        seq = Sequence()
        seq.read(file_path=str(self.seq_path))
        k_traj_adc_numpy = seq.calculate_kspace()[0]
        k_traj_adc = torch.tensor(k_traj_adc_numpy, dtype=torch.float32)
        k_traj_reshaped = repeat(k_traj_adc, 'xyz (other k0) -> xyz other coils k2 k1 k0', coils=1, k2=1, k1=1, k0=n_k0)

        return KTrajectory.from_tensor(
            k_traj_reshaped,
            axes_order='xyz',
            scaling_matrix=encoding_matrix,
            repeat_detection_tolerance=self.repeat_detection_tolerance,
        )



"""Class for Fast Fourier Operator."""

from collections.abc import Sequence
from dataclasses import astuple

import torch

from mrpro.data.SpatialDimension import SpatialDimension
from mrpro.operators.LinearOperator import LinearOperator
from mrpro.operators.ZeroPadOp import ZeroPadOp


class FastFourierOp(LinearOperator):
    """Fast Fourier operator class.

    Applies a Fast Fourier Transformation along selected dimensions with cropping/zero-padding
    along these selected dimensions

    The transformation is done with ``'ortho'`` normalization, i.e. the normalization constant is split between
    forward and adjoint [FFT]_.

    .. note::
        The input to both `~FastFourierOp.forward` and `~FastFourierOp.adjoint`
        are assumed to have the zero-frequency in the center of the data. `torch.fft.fftn`
        and `torch.fft.ifftn` expect the zero-frequency to be the first entry in the tensor.
        Therefore in `~FastFourierOp.forward` and `~FastFourierOp.adjoint`,
        first `torch.fft.ifftshift`, then `torch.fft.fftn` or `torch.fft.ifftn`,
        finally `torch.fft.fftshift` are applied.

    .. note::
       See also `~mrpro.operators.FourierOp` for a Fourier operator that handles
       automatic sorting of the k-space data based on a trajectory.


    References
    ----------
    .. [FFT] https://numpy.org/doc/stable/reference/routines.fft.html

    """

    def __init__(
        self,
        dim: Sequence[int] = (-3, -2, -1),
        recon_matrix: SpatialDimension[int] | Sequence[int] | None = None,
        encoding_matrix: SpatialDimension[int] | Sequence[int] | None = None,
    ) -> None:
        """Initialize a Fast Fourier Operator.

        If both `recon_matrix` and `encoding_matrix` are set, the operator will perform padding/cropping before and
        after the transforms to match the shape in image space (`recon_matrix`) and k-shape (`encoding_matrix`).
        If both are set to `None`, no padding or cropping will be performed.
        If these are `~mrpro.data.SpatialDimension`, the transform dimensions must be within the last three dimensions,
        typically corresponding to the `(k2, k1, k0)` and `(z, y, x)` axes of `~mrpro.data.KData`
        and `~mrpro.data.IData`, respectively.


        Parameters
        ----------
        dim
            dim along which FFT and IFFT are applied, by default last three dimensions,
            as these correspond to `k2`, `k1`, and `k0` of k-space data.
        encoding_matrix
            shape of encoded k-data along the axes in `dim`. Must be set if `recon_matrix` is set.
            If `encoding_matrix` and `recon_matrix` are `None`, no padding or cropping will be performed.
            If all values in dim are -3, -2 or -1, this can also be a `~mrpro.data.SpatialDimension` describing the
            k-space shape in all 3 dimensions `(k2, k1, k0)`, but only values in the dimensions in `dim` will be used.
            Otherwise, it should be a `Sequence` of the same length as `dim`.
        recon_matrix
            shape of reconstructed image data. Must be set if `encoding_matrix` is set.
            If `encoding_matrix` and `recon_matrix` are `None`, no padding or cropping will be performed.
            If all values in `dim` are -3, -2 or -1, this can also be a `~mrpro.data.SpatialDimension` describing the
            image-space shape in all 3 dimensions `(z, y, x)`, but only values in the dimensions in `dim` will be used.
            Otherwise, it should be a `Sequence` of the same length as `dim`.
        """
        super().__init__()
        self._dim = tuple(dim)
        self._pad_op: ZeroPadOp

        if isinstance(recon_matrix, SpatialDimension):
            if not all(d in (-1, -2, -3) for d in dim):
                raise NotImplementedError(
                    f'recon_matrix can only be a SpatialDimension if each value in dim is in (-3,-2,-1),'
                    f'got {dim=}\nInstead, you can also supply a list of values of same length as dim'
                )
            original_shape: Sequence[int] | None = [int(astuple(recon_matrix)[d]) for d in dim]

        else:
            original_shape = recon_matrix

        if isinstance(encoding_matrix, SpatialDimension):
            if not all(d in (-1, -2, -3) for d in dim):
                raise NotImplementedError(
                    f'encoding_matrix can only be a SpatialDimension if each value in dim is in (-3,-2,-1),'
                    f'got {dim=}\nInstead, you can also supply a list of values of same length as dim'
                )
            padded_shape: Sequence[int] | None = [int(astuple(encoding_matrix)[d]) for d in dim]

        else:
            padded_shape = encoding_matrix

        if original_shape is not None and padded_shape is not None:
            # perform padding / cropping
            self._pad_op = ZeroPadOp(dim=dim, original_shape=original_shape, padded_shape=padded_shape)
        elif encoding_matrix is None and recon_matrix is None:
            # No padding no padding / cropping
            self._pad_op = ZeroPadOp(dim=(), original_shape=(), padded_shape=())
        else:
            raise ValueError(
                'Either encoding_matrix and recon_matrix must both be set to None or both to a value, got'
                f'{encoding_matrix=} and {recon_matrix=}'
            )

    def __call__(self, x: torch.Tensor) -> tuple[torch.Tensor,]:
        """Apply Fast Fourier Transform (FFT) from image space to k-space.

        Parameters
        ----------
        x
            (image-space) data on a Cartesian grid.

        Returns
        -------
            FFT of `x`
        """
        return super().__call__(x)

    def forward(self, x: torch.Tensor) -> tuple[torch.Tensor,]:
        """Apply forward of FastFourierOp.

        .. note::
            Prefer calling the instance of the FastFourierOp operator as ``operator(x)`` over
            directly calling this method. See this PyTorch `discussion <https://discuss.pytorch.org/t/is-model-forward-x-the-same-as-model-call-x/33460/3>`_.
        """
        y = torch.fft.fftshift(
            torch.fft.fftn(torch.fft.ifftshift(*self._pad_op(x), dim=self._dim), dim=self._dim, norm='ortho'),
            dim=self._dim,
        )
        return (y,)

    def adjoint(self, y: torch.Tensor) -> tuple[torch.Tensor,]:
        """Apply Inverse Fast Fourier Transform (IFFT) from k-space to image space.

        Parameters
        ----------
        y
            (k-space) data on a Cartesian grid.

        Returns
        -------
            IFFT of `y`
        """
        # FFT
        return self._pad_op.adjoint(
            torch.fft.fftshift(
                torch.fft.ifftn(torch.fft.ifftshift(y, dim=self._dim), dim=self._dim, norm='ortho'),
                dim=self._dim,
            ),
        )

    def __repr__(self) -> str:
        """Representation method for FastFourierOperator."""
        k2k1k0 = ['k2', 'k1', 'k0']
        dimension_210 = tuple(k2k1k0[i] for i in self._dim if i in range(-3, 0))
        dimension_210_str = ', '.join(dimension_210)
        out = f'Dimension(s) along which FFT is applied: ({dimension_210_str})'
        return out



"""Primal-Dual Hybrid Gradient Algorithm (PDHG)."""

from __future__ import annotations

import warnings
from collections.abc import Callable, Sequence
from dataclasses import dataclass

import torch

from mrpro.algorithms.optimizers.OptimizerStatus import OptimizerStatus
from mrpro.operators import (
    IdentityOp,
    LinearOperator,
    LinearOperatorMatrix,
    ProximableFunctional,
    ProximableFunctionalSeparableSum,
)
from mrpro.operators.functionals import ZeroFunctional


@dataclass
class PDHGStatus(OptimizerStatus):
    """Status of the PDHG algorithm."""

    objective: Callable[[*tuple[torch.Tensor, ...]], torch.Tensor]
    dual_stepsize: float | torch.Tensor
    primal_stepsize: float | torch.Tensor
    relaxation: float | torch.Tensor
    duals: Sequence[torch.Tensor]
    relaxed: Sequence[torch.Tensor]


def _norm_squared(*values: torch.Tensor) -> torch.Tensor:
    """Calculate the squared L2 norm of the stack of tensors."""
    return sum([torch.vdot(value.flatten(), value.flatten()).real for value in values], start=torch.tensor(0.0))


def pdhg(
    f: ProximableFunctionalSeparableSum | ProximableFunctional | None,
    g: ProximableFunctionalSeparableSum | ProximableFunctional | None,
    operator: LinearOperator | LinearOperatorMatrix | None,
    initial_values: Sequence[torch.Tensor],
    max_iterations: int = 32,
    tolerance: float = 0,
    primal_stepsize: float | torch.Tensor | None = None,
    dual_stepsize: float | torch.Tensor | None = None,
    relaxation: float = 1.0,
    initial_relaxed: Sequence[torch.Tensor] | None = None,
    initial_duals: Sequence[torch.Tensor] | None = None,
    callback: Callable[[PDHGStatus], bool | None] | None = None,
) -> tuple[torch.Tensor, ...]:
    r"""Primal-Dual Hybrid Gradient Algorithm (PDHG).

    Solves the minimization problem

        :math:`\min_x f(K x) + g(x)`

    with linear operator :math:`K` and proper, convex, lower-semicontinous functionals :math:`f` and :math:`g`.

    PDHG is a primal-dual algorithm that performs the following steps

    .. math::

        z_{k+1} = \mathrm{prox}_{\sigma f^{\ast}}(z_k + \sigma K \bar{x}_k),
        x_{k+1} = \mathrm{prox}_{\tau g}(x_k - \tau K^H z_{k+1}),
        \bar{x}_{k+1} = x_{k+1} + \theta(x_{k+1} - x_k),

    where :math:`\mathrm{prox}` denotes the proximal operator and :math:`f^{\ast}` is the
    convex conjugate of the functional :math:`f`. Thereby, :math:`\tau` and :math:`\sigma`
    are the primal and dual step sizes, respectively (see further below) and :math:`\theta\in [0,1]`.

    The operator can be supplied as a `~mrpro.operators.LinearOperator` or as a
    :math:`m\times n` -`~mrpro.operators.LinearOperatorMatrix`
    (you can use ``A | B`` for horizontal, ``A % B`` for vertical stacking).
    :math:`f` and :math:`g` can be single functionals or a `~mrpro.operators.ProximableFunctionalSeparableSum` of m or n
    functionals, respectively (build with ``f | g``).

    Thus, this implementation solves the problem

        :math:`\min_{x=(x_1,\ldots,x_n)} \sum_{i=1}^m f_i\big( (Kx)_i\big) + \sum_{j=1}^n g_j(x_j)`.

    If neither primal nor dual step size are supplied, they are both chosen as :math:`1/||K||_2`.
    If only one of them is supplied, the other is chosen such that

        :math:`\tau \sigma = 1/||K||_2`,

    where :math:`1/||K||_2` denotes the operator-norm of :math:`K`.
    Note that the computation of the operator-norm can be computationally expensive and
    that if no stepsizes are provided, the algorithm runs a power iteration to obtain the
    upper bound for the stepsizes.

    For a warm start, the initial relaxed primal and dual variables can be supplied.
    These might be obtained from the status object of a previous run.

    Parameters
    ----------
    f
        Functional `f` in the problem definition. If set to None, it is interpreted as the zero-functional.
    g
        Functional `g` in the problem definition. If set to None, it is interpreted as the zero-functional.
    operator
        Linear operator or matrix of linear operators;
        if set to `None`, it is interpreted as the Identity-operator.
    initial_values
        initial guess of the solution.
    max_iterations
        maximum number of iterations.
    tolerance
        tolerance for relative change of the primal solution; if set to zero, `max_iterations` of pdhg are run.
    dual_stepsize
        dual step size.
    primal_stepsize
        primal step size.
    relaxation
        relaxation parameter, ``1.0`` is no relaxation.
    initial_relaxed
        relaxed primals, used for warm start.
    initial_duals
        dual variables, used for warm start.
    callback
        callback function called after each iteration. This can be used to monitor the progress of the algorithm.
        If it returns `False`, the algorithm stops at that iteration.
    """
    if f is None and g is None:
        warnings.warn(
            'Both f and g are None. The objective is constant. Returning x0 as a possible solution', stacklevel=2
        )
        return tuple(initial_values)

    if operator is None:
        # Use identity operator if no operator is supplied
        n_rows = len(f) if isinstance(f, ProximableFunctionalSeparableSum) else 1
        n_columns = len(g) if isinstance(g, ProximableFunctionalSeparableSum) else 1
        if n_rows != n_columns:
            raise ValueError('If operator is None, the number of elements in f and g should be the same')
        operator_matrix = LinearOperatorMatrix.from_diagonal(*((IdentityOp(),) * n_rows))
    else:
        if isinstance(operator, LinearOperator):
            # We always use a matrix of operators for homogeneous handling
            operator_matrix = LinearOperatorMatrix.from_diagonal(operator)
        else:
            operator_matrix = operator
        n_rows, n_columns = operator_matrix.shape

    # We always use a separable sum for homogeneous handling, even if it is just a ZeroFunctional
    if f is None:
        f_sum: ProximableFunctionalSeparableSum = ProximableFunctionalSeparableSum(*(ZeroFunctional(),) * n_rows)
    elif isinstance(f, ProximableFunctional):
        f_sum = ProximableFunctionalSeparableSum(f)
    else:
        f_sum = f

    if len(f_sum) != n_rows:
        raise ValueError('Number of rows in operator does not match number of functionals in f')

    if g is None:
        g_sum: ProximableFunctionalSeparableSum = ProximableFunctionalSeparableSum(*(ZeroFunctional(),) * n_columns)
    elif isinstance(g, ProximableFunctional):
        g_sum = ProximableFunctionalSeparableSum(g)
    else:
        g_sum = g

    if len(g_sum) != n_columns:
        raise ValueError('Number of columns in operator does not match number of functionals in g')

    if primal_stepsize is None or dual_stepsize is None:
        # choose primal and dual step size such that their product is 1/|operator|**2
        # to ensure convergence
        operator_norm = operator_matrix.operator_norm(*[torch.randn_like(v) for v in initial_values]).amax()
        if primal_stepsize is None and dual_stepsize is None:
            primal_stepsize_ = dual_stepsize_ = 1.0 / operator_norm
        elif primal_stepsize is None and dual_stepsize is not None:
            primal_stepsize_ = 1 / (operator_norm * dual_stepsize)
            dual_stepsize_ = torch.as_tensor(dual_stepsize)
        elif dual_stepsize is None and primal_stepsize is not None:
            dual_stepsize_ = 1 / (operator_norm * primal_stepsize)
            primal_stepsize_ = torch.as_tensor(primal_stepsize)
    else:
        primal_stepsize_ = torch.as_tensor(primal_stepsize)
        dual_stepsize_ = torch.as_tensor(dual_stepsize)

    primals_relaxed = initial_values if initial_relaxed is None else initial_relaxed
    duals = (0 * operator_matrix)(*initial_values) if initial_duals is None else initial_duals

    if len(duals) != n_rows:
        raise ValueError('if dual variable is supplied, it should be a tuple of same length as the tuple of g')

    primals = initial_values
    for i in range(max_iterations):
        duals = tuple(
            dual + dual_stepsize_ * step for dual, step in zip(duals, operator_matrix(*primals_relaxed), strict=False)
        )
        duals = f_sum.prox_convex_conj(*duals, sigma=dual_stepsize_)

        primals_new = tuple(
            primal - primal_stepsize_ * step for primal, step in zip(primals, operator_matrix.H(*duals), strict=False)
        )
        primals_new = g_sum.prox(*primals_new, sigma=primal_stepsize_)
        primals_relaxed = tuple(
            torch.lerp(primal, primal_new, relaxation) for primal, primal_new in zip(primals, primals_new, strict=False)
        )

        # check if the solution is already accurate enough
        if tolerance != 0:
            change_squared = _norm_squared(*[(old - new) for old, new in zip(primals, primals_new, strict=True)])
            primals_new_squared = _norm_squared(*primals_new)
            if change_squared < tolerance**2 * primals_new_squared:
                return tuple(primals_new)

        primals = primals_new

        if callback is not None:
            status = PDHGStatus(
                iteration_number=i,
                dual_stepsize=dual_stepsize_,
                primal_stepsize=primal_stepsize_,
                relaxation=relaxation,
                duals=duals,
                solution=tuple(primals),
                relaxed=primals_relaxed,
                objective=lambda *x: f_sum.forward(*operator_matrix(*x))[0] + g_sum.forward(*x)[0],
            )
            continue_iterations = callback(status)
            if continue_iterations is False:
                break

    return tuple(primals)



"""MR image data header (IHeader) dataclass."""

import dataclasses
import datetime as dt
import re
from collections.abc import Callable, Sequence
from dataclasses import field
from datetime import datetime, time
from typing import cast

import numpy as np
import torch
from einops import rearrange
from pydicom.dataset import Dataset
from pydicom.tag import Tag, TagType
from typing_extensions import Self, TypeVar

from mrpro.data.AcqInfo import AcqIdx, PhysiologyTimestamps
from mrpro.data.Dataclass import Dataclass
from mrpro.data.KHeader import KHeader
from mrpro.data.Rotation import Rotation
from mrpro.data.SpatialDimension import SpatialDimension
from mrpro.utils.reduce_repeat import reduce_repeat
from mrpro.utils.reshape import unsqueeze_right
from mrpro.utils.unit_conversion import deg_to_rad, mm_to_m, ms_to_s


def _int_factory() -> torch.Tensor:
    return torch.zeros(1, 1, 1, 1, 1, dtype=torch.int64)  # other, coil, z, y, x


def parse_datetime(datetime_str: str) -> datetime:
    """Parse datetime string with or without timezone information.

    Parameters
    ----------
    datetime_str
        Datetime string which may has timezone information.

    Returns
    -------
        Datetime
    """
    # Check if the string has a timezone (e.g., "+0000", "-0800", "Z")
    if re.search(r'([+-]\d{2}:?\d{2}|Z)$', datetime_str):
        dt = datetime.strptime(datetime_str, '%Y%m%d%H%M%S.%f%z')
        return dt.replace(tzinfo=None)  # Remove timezone info
    else:
        return datetime.strptime(datetime_str, '%Y%m%d%H%M%S.%f')


T = TypeVar('T')


def get_items(datasets: Sequence[Dataset], name: TagType, target_type: Callable[..., T]) -> list[T]:
    """Get a flattened list of converted items from each dataset for a given name or Tag.

    Parameters
    ----------
    name
        The name or tag to filter items.
    datasets
        The datasets to extract the items from.
    target_type
        The target type to convert the items into.

    Returns
    -------
        A list of converted items.
    """
    return [
        target_type(item.value)
        for ds in datasets
        for item in ds.iterall()
        if (item.tag == Tag(name)) and item.value is not None
    ]


def try_reduce_repeat(value: T) -> T:
    """Replace dimensions by singleton if possible, raise exception if spatial or coil dimension remains."""
    # TODO: Should be replaced by the CheckDataMixin and ReduceRepeatMixin
    match value:
        case torch.Tensor():
            tensor = reduce_repeat(value, 1e-6)
            if tensor.shape[-4:].numel() == 1:
                return cast(T, tensor)
        case Rotation():
            tensor = value.as_matrix()
            tensor = reduce_repeat(tensor, 1e-4, range(tensor.ndim - 1))
            if tensor.shape[-5:-2].numel() == 1:
                return cast(T, Rotation.from_matrix(tensor))
        case SpatialDimension():
            spatialdimension = value.apply(lambda x: reduce_repeat(x, 1e-6))
            if spatialdimension.shape[-4:].numel() == 1:
                return cast(T, spatialdimension)
        case list():
            return cast(T, list(value))
        case _:
            raise NotImplementedError('Unsupported Type')

    raise ValueError(f'Dimension mismatch. Spatial or coil dimension should be reduced to a single value. {value}')


class ImageIdx(Dataclass):
    """Indices for each slice or volume.

    The different counters describe the use of each slice or volume in the image data.
    See MRD Image specification for more information.

    References
    ----------
    .. [1] MRD Image Specification
       https://ismrmrd.readthedocs.io/en/latest/mrd_image_data.html
    """

    average: torch.Tensor = field(default_factory=_int_factory)
    """Signal average."""

    slice: torch.Tensor = field(default_factory=_int_factory)
    """Slice number (multi-slice 2D)."""

    contrast: torch.Tensor = field(default_factory=_int_factory)
    """Echo number in multi-echo."""

    phase: torch.Tensor = field(default_factory=_int_factory)
    """Cardiac phase."""

    repetition: torch.Tensor = field(default_factory=_int_factory)
    """Counter in repeated/dynamic acquisitions."""

    set: torch.Tensor = field(default_factory=_int_factory)
    """Sets of different preparation, e.g. flow encoding, diffusion weighting."""

    user0: torch.Tensor = field(default_factory=_int_factory)
    """User index 0."""

    user1: torch.Tensor = field(default_factory=_int_factory)
    """User index 1."""

    user2: torch.Tensor = field(default_factory=_int_factory)
    """User index 2."""

    user3: torch.Tensor = field(default_factory=_int_factory)
    """User index 3."""

    user4: torch.Tensor = field(default_factory=_int_factory)
    """User index 4."""

    user5: torch.Tensor = field(default_factory=_int_factory)
    """User index 5."""

    user6: torch.Tensor = field(default_factory=_int_factory)
    """User index 6."""

    user7: torch.Tensor = field(default_factory=_int_factory)
    """User index 7."""

    @classmethod
    def from_acqidx(cls, idx: AcqIdx) -> Self:
        """Create ImageIdx object from AcqIdx object.

        This copies the indices except k1 and k2, which are not used in the image header.

        Parameters
        ----------
        idx
            Acquisition indices.
        """
        return cls(
            average=try_reduce_repeat(idx.average),
            slice=try_reduce_repeat(idx.slice),
            contrast=try_reduce_repeat(idx.contrast),
            phase=try_reduce_repeat(idx.phase),
            repetition=try_reduce_repeat(idx.repetition),
            set=try_reduce_repeat(idx.set),
            user0=try_reduce_repeat(idx.user0),
            user1=try_reduce_repeat(idx.user1),
            user2=try_reduce_repeat(idx.user2),
            user3=try_reduce_repeat(idx.user3),
            user4=try_reduce_repeat(idx.user4),
            user5=try_reduce_repeat(idx.user5),
            user6=try_reduce_repeat(idx.user6),
            user7=try_reduce_repeat(idx.user7),
        )


class IHeader(Dataclass):
    """MR image data header."""

    resolution: SpatialDimension[float]
    """Pixel spacing [m/px]."""

    te: list[float] | torch.Tensor = field(default_factory=list)
    """Echo time [s]."""

    ti: list[float] | torch.Tensor = field(default_factory=list)
    """Inversion time [s]."""

    fa: list[float] | torch.Tensor = field(default_factory=list)
    """Flip angle [rad]."""

    tr: list[float] | torch.Tensor = field(default_factory=list)
    """Repetition time [s]."""

    _misc: dict = dataclasses.field(default_factory=dict)
    """Dictionary with miscellaneous parameters."""

    position: SpatialDimension[torch.Tensor] = field(
        default_factory=lambda: SpatialDimension(  # other, coil, z, y, x
            torch.zeros(1, 1, 1, 1, 1),
            torch.zeros(1, 1, 1, 1, 1),
            torch.zeros(1, 1, 1, 1, 1),
        )
    )
    """Center of the image or volume [m]"""

    orientation: Rotation = field(default_factory=lambda: Rotation.identity((1, 1, 1, 1, 1)))
    """Orientation of the image or volume"""

    patient_table_position: SpatialDimension[torch.Tensor] = field(
        default_factory=lambda: SpatialDimension(  # other, coil, z, y, x
            torch.zeros(1, 1, 1, 1, 1),
            torch.zeros(1, 1, 1, 1, 1),
            torch.zeros(1, 1, 1, 1, 1),
        )
    )
    """Offset position of the patient table [m]"""

    acquisition_time_stamp: torch.Tensor = field(default_factory=lambda: torch.zeros(1, 1, 1, 1, 1))
    """Clock time stamp of the slice or volume [s]"""

    physiology_time_stamps: PhysiologyTimestamps = field(default_factory=PhysiologyTimestamps)
    """Time stamps relative to physiological triggering, e.g. ECG [s]."""

    idx: ImageIdx = field(default_factory=ImageIdx)
    """Image Counters. For each slice or volume, describe its use."""

    datetime: dt.datetime | None = None
    """Date and time the image."""

    @classmethod
    def from_kheader(cls, header: KHeader) -> Self:
        """Create IHeader object from KHeader object.

        Parameters
        ----------
        header
            MR raw data header (KHeader) containing required meta data.
        """
        resolution = header.recon_fov / header.recon_matrix

        # TODO: how to deal with different values for each acquisition?
        return cls(
            resolution=resolution,
            te=try_reduce_repeat(header.te),
            ti=try_reduce_repeat(header.ti),
            fa=try_reduce_repeat(header.fa),
            tr=try_reduce_repeat(header.tr),
            orientation=try_reduce_repeat(header.acq_info.orientation),
            position=try_reduce_repeat(header.acq_info.position.apply(lambda x: x)),
            patient_table_position=try_reduce_repeat(header.acq_info.patient_table_position.apply(lambda x: x)),
            acquisition_time_stamp=try_reduce_repeat(header.acq_info.acquisition_time_stamp.mean((-1, -2, -3), True)),
            physiology_time_stamps=PhysiologyTimestamps(
                try_reduce_repeat(header.acq_info.physiology_time_stamps.timestamp0.mean((-1, -2, -3), True)),
                try_reduce_repeat(header.acq_info.physiology_time_stamps.timestamp1.mean((-1, -2, -3), True)),
                try_reduce_repeat(header.acq_info.physiology_time_stamps.timestamp2.mean((-1, -2, -3), True)),
            ),
            idx=ImageIdx.from_acqidx(header.acq_info.idx),
            datetime=header.datetime,
        )

    @classmethod
    def from_dicom(cls, *dataset: Dataset) -> Self:
        """Read DICOM files and return IHeader object.

        Parameters
        ----------
        dataset
            one or multiple dataset objects containing the DICOM data.
        """
        # Ensure datasets are consistently single-frame or multi-frame 2D/3D
        number_of_frames = get_items(dataset, 'NumberOfFrames', int)
        if len(set(number_of_frames)) > 1:
            raise ValueError('Only DICOM files with the same number of frames can be stacked.')

        mr_acquisition_type = get_items(dataset, 'MRAcquisitionType', str)
        if len(set(mr_acquisition_type)) > 1:
            raise ValueError('Only DICOM files with the same MRAcquisitionType can be stacked.')

        pixel_spacing = get_items(dataset, 'PixelSpacing', lambda x: [float(val) for val in x])
        if not pixel_spacing or len(np.unique(torch.tensor(pixel_spacing), axis=0)) > 1:
            raise ValueError('Pixel spacing needs to be defined and the same for all.')

        slice_thickness = get_items(dataset, 'SliceThickness', float)
        if len(set(slice_thickness)) != 1:
            raise ValueError('Slice thickness needs to be defined and the same for all.')

        # The only mandatory field
        resolution = SpatialDimension(
            z=slice_thickness[0],
            y=pixel_spacing[0][1],
            x=pixel_spacing[0][0],
        ).apply_(mm_to_m)

        is_3d = number_of_frames and number_of_frames[0] > 1 and mr_acquisition_type and mr_acquisition_type[0] == '3D'
        num_slices = number_of_frames[0] if number_of_frames and is_3d else 1

        header = cls(resolution=resolution)

        # For 3D datasets we currently want to save only one value per volume of fa, ti, tr, and te
        if fa := get_items(dataset, 'FlipAngle', float)[::num_slices]:
            header.fa = deg_to_rad(try_reduce_repeat(fa))
        if ti := get_items(dataset, 'InversionTime', float)[::num_slices]:
            header.ti = ms_to_s(try_reduce_repeat(ti))
        if tr := get_items(dataset, 'RepetitionTime', float)[::num_slices]:
            header.tr = ms_to_s(try_reduce_repeat(tr))
        # Some scanners use 'EchoTime', some use 'EffectiveEchoTime'
        if not (te_ms := get_items(dataset, 'EchoTime', float)[::num_slices]):
            te_ms = get_items(dataset, 'EffectiveEchoTime', float)[::num_slices]
        if te_ms:
            header.te = ms_to_s(try_reduce_repeat(te_ms))

        # Dicom orientation is described by two vectors in the format (x1, y1, z1, x2, y2, z2)
        if dcm_orientation := get_items(dataset, 'ImageOrientationPatient', lambda x: [np.float32(val) for val in x]):
            phase_direction = torch.tensor(dcm_orientation[0][:3])  # contains (x1, y1, z1)
            readout_direction = torch.tensor(dcm_orientation[0][3:])  # contains (x2, y2, z2)
            # Calculate third basis vector by cross product
            slice_direction = torch.cross(readout_direction, phase_direction, dim=-1)  # contains (x3, y3, z3)
            # Get orientation from basis vectors, mrpro order: z, y, x
            header.orientation = Rotation.from_directions(
                SpatialDimension.from_array_xyz(slice_direction),
                SpatialDimension.from_array_xyz(phase_direction),
                SpatialDimension.from_array_xyz(readout_direction),
            )

        # Get dicom position: Pixel with index (0, 0, 0) for (x, y, z), i.e. voxel center in the upper left corner
        if dcm_position := SpatialDimension.from_array_xyz(
            get_items(dataset, 'ImagePositionPatient', lambda x: [np.float32(val) for val in x]),
        ):
            # Get shift and apply orientation
            n_rows = get_items(dataset, 'Rows', int)[0]
            n_cols = get_items(dataset, 'Columns', int)[0]
            shift_rotated = SpatialDimension.from_array_xyz(torch.zeros(3))
            if n_rows and n_cols:
                # Get shift in [m]
                shift = resolution * SpatialDimension(x=n_rows, y=n_cols, z=0) / 2
                # Get the shift vector for image position by applying the fov rotation
                shift_rotated = header.orientation(shift)

            # Get dicom image position in [m] and apply shift defined in [m]
            dcm_position = dcm_position.rearrange('(other z) -> other 1 z 1 1', z=num_slices)
            position = dcm_position.apply(mm_to_m) + shift_rotated
            if is_3d and num_slices > 1:
                z_shift = (
                    header.orientation.as_directions()[0] * (-num_slices / 2 + torch.arange(num_slices)) * resolution
                )
                z_shift = z_shift.rearrange('z -> 1 1 z 1 1')
                position = position - z_shift
            header.position = position

        # Calculate acquisition time stamps in s according to the reference time 0:00 am
        if frame_time_dt := get_items(dataset, 'FrameReferenceDateTime', parse_datetime)[::num_slices]:
            t0 = datetime.combine(frame_time_dt[0].date(), time(0, 0))
            time_stamp = torch.tensor([(ft - t0).total_seconds() for ft in frame_time_dt])
            header.acquisition_time_stamp = unsqueeze_right(time_stamp, 4)

        if dcm_cardiac_trigger := get_items(dataset, 'NominalCardiacTriggerDelayTime', float):
            header.physiology_time_stamps = PhysiologyTimestamps(
                timestamp1=unsqueeze_right(ms_to_s(torch.tensor(dcm_cardiac_trigger)), 4)
            )

        # The in stack position accounts for the slice position for multi-file data with cardiac phases,
        # as well as for single file dicom with multiple slices. Index is reduced by 1 to start indexing at 0.
        if phase_idx := get_items(dataset, 'TemporalPositionIndex', int):
            phase_idx_tensor = torch.tensor(try_reduce_repeat(phase_idx)) - 1
            header.idx.phase = rearrange(phase_idx_tensor, '(other z) -> other 1 z 1 1', z=num_slices)
        if slice_idx := get_items(dataset, 'InStackPositionNumber', int):
            slice_idx_tensor = torch.tensor(try_reduce_repeat(slice_idx)) - 1
            header.idx.slice = rearrange(slice_idx_tensor, '(other z) -> other 1 z 1 1', z=num_slices)

        return header



"""Pre-built reconstruction algorithms."""

from mrpro.algorithms.reconstruction.Reconstruction import Reconstruction
from mrpro.algorithms.reconstruction.DirectReconstruction import DirectReconstruction
from mrpro.algorithms.reconstruction.RegularizedIterativeSENSEReconstruction import RegularizedIterativeSENSEReconstruction
from mrpro.algorithms.reconstruction.IterativeSENSEReconstruction import IterativeSENSEReconstruction
from mrpro.algorithms.reconstruction.TotalVariationRegularizedReconstruction import TotalVariationRegularizedReconstruction
__all__ = [
    "DirectReconstruction",
    "IterativeSENSEReconstruction",
    "Reconstruction",
    "RegularizedIterativeSENSEReconstruction",
    "TotalVariationRegularizedReconstruction"
]



"""Class for Sensitivity Operator."""

import torch

from mrpro.data.CsmData import CsmData
from mrpro.operators.LinearOperator import LinearOperator


class SensitivityOp(LinearOperator):
    """Sensitivity operator class.

    The forward operator expands an image to multiple coil images according to coil sensitivity maps,
    the adjoint operator reduces the coil images to a single image.
    """

    def __init__(self, csm: CsmData | torch.Tensor) -> None:
        """Initialize a Sensitivity Operator.

        Parameters
        ----------
        csm
           Coil Sensitivity Data
        """
        super().__init__()
        if isinstance(csm, CsmData):
            # only tensors can be used as buffers
            csm = csm.data
        self.csm_tensor = csm

    def __call__(self, img: torch.Tensor) -> tuple[torch.Tensor,]:
        """Apply coil sensitivity maps to expand the coil dimension.

        This operator multiplies the input image (assumed to be a single coil image
        or a combined image) by the coil sensitivity maps (CSM) to produce
        multi-coil image data.

        Parameters
        ----------
        img
            Input image data, typically with shape `(... 1 z y x)`.

        Returns
        -------
            Multi-coil image data with shape `(... coils z y x)`.
        """
        return super().__call__(img)

    def forward(self, img: torch.Tensor) -> tuple[torch.Tensor,]:
        """Apply forward of SensitivityOp.

        .. note::
            Prefer calling the instance of the SensitivityOp operator as ``operator(x)`` over
            directly calling this method. See this PyTorch `discussion <https://discuss.pytorch.org/t/is-model-forward-x-the-same-as-model-call-x/33460/3>`_.
        """
        return (self.csm_tensor * img,)

    def adjoint(self, img: torch.Tensor) -> tuple[torch.Tensor,]:
        """Apply adjoint sensitivity mapping to reduce the coil dimension.

        This operator multiplies the input multi-coil image data by the complex
        conjugate of the coil sensitivity maps (CSM) and then sums along
        the coil dimension, effectively performing a coil combination.

        Parameters
        ----------
        img
            Multi-coil image data, typically with shape `(... coils z y x)`.

        Returns
        -------
            Combined image data, with shape `(... 1 z y x)`.
        """
        return ((self.csm_tensor.conj() * img).sum(-4, keepdim=True),)



"""(Iterative) Walsh method for coil sensitivity map calculation."""

import torch

from mrpro.data.SpatialDimension import SpatialDimension
from mrpro.utils.filters import uniform_filter


def walsh(coil_images: torch.Tensor, smoothing_width: SpatialDimension[int] | int) -> torch.Tensor:
    """Calculate a coil sensitivity map (csm) using an iterative version of the Walsh method.

    This function computes CSMs from a set of complex coil images assuming spatially
    slowly changing sensitivity maps using Walsh's method [WAL2000]_.

    The algorithm follows these steps:

    1. **Compute Pointwise Covariance**:
       Calculate the covariance matrix of the coil images at each voxel to capture inter-coil signal relationships.

    2. **Apply Smoothing Filter**:
       Smooth the covariance matrices across spatial dimensions using a uniform filter of specified width
       to reduce noise and enforce spatial consistency.

    3. **Dominant Eigenvector Estimation via Power Iteration**:
       Perform power iterations to approximate the dominant eigenvector of the covariance matrix at each voxel,
       representing the principal component of the signal.

    4. **Normalize Sensitivity Maps**:
       Normalize the resulting eigenvectors to produce the final CSMs.

    This function works on a single set of coil images. The input should be a tensor with dimensions
    `(coils, z, y, x)`. The output will have the same dimensions. Either apply this function individually to each set of
    coil images, or see `~mrpro.data.CsmData.from_idata_walsh` which performs this operation on a whole dataset
    [WAL2000]_.

    This implementation is inspired by `ismrmrd-python-tools <https://github.com/ismrmrd/ismrmrd-python-tools>`_.

    Parameters
    ----------
    coil_images
        images for each coil element
    smoothing_width
        width of the smoothing filter

    References
    ----------
    .. [WAL2000] Walsh DO, Gmitro AF, Marcellin MW (2000) Adaptive reconstruction of phased array MR imagery. MRM 43
    """
    # After 10 power iterations we will have a very good estimate of the singular vector
    n_power_iterations = 10

    if isinstance(smoothing_width, int):
        smoothing_width = SpatialDimension(smoothing_width, smoothing_width, smoothing_width)
    # Compute the pointwise covariance between coils
    coil_covariance = torch.einsum('azyx,bzyx->abzyx', coil_images, coil_images.conj())

    # Smooth the covariance along y-x for 2D and z-y-x for 3D data
    coil_covariance = uniform_filter(coil_covariance, width=smoothing_width.zyx, dim=(-3, -2, -1))

    # At each point in the image, find the dominant eigenvector
    # of the signal covariance matrix using the power method
    v = coil_covariance.sum(dim=0)
    for _ in range(n_power_iterations):
        v /= v.norm(dim=0)
        v = torch.einsum('abzyx,bzyx->azyx', coil_covariance, v)
    csm = v / v.norm(dim=0)

    # Make sure there are no inf or nan-values due to very small values in the covariance matrix
    # nan_to_num does not work for complexfloat, boolean indexing not with vmap.
    csm = torch.where(torch.isfinite(csm), csm, 0.0)
    return csm



"""Optimizers."""

from mrpro.algorithms.optimizers.OptimizerStatus import OptimizerStatus
from mrpro.algorithms.optimizers.adam import adam
from mrpro.algorithms.optimizers.cg import cg
from mrpro.algorithms.optimizers.lbfgs import lbfgs
from mrpro.algorithms.optimizers.pdhg import pdhg
from mrpro.algorithms.optimizers.pgd import pgd
__all__ = ["OptimizerStatus", "adam", "cg", "lbfgs", "pdhg", "pgd"]



"""Separable Sum of Proximable Functionals."""

from __future__ import annotations

import operator
from collections.abc import Iterator
from functools import reduce
from typing import cast

import torch
from typing_extensions import TypeVarTuple, Unpack, overload

from mrpro.operators.Functional import ProximableFunctional
from mrpro.operators.Operator import Operator

T = TypeVarTuple('T')


class ProximableFunctionalSeparableSum(Operator[Unpack[T], tuple[torch.Tensor]]):
    r"""Separable Sum of Proximable Functionals.

    This is a separable sum of the functionals. The forward method returns the sum of the functionals
    evaluated at the inputs, :math:`\sum_i f_i(x_i)`.
    Use ``f | g`` to build a separable sum from two proximable functionals.
    """

    functionals: tuple[ProximableFunctional, ...]

    @overload
    def __init__(self: ProximableFunctionalSeparableSum[torch.Tensor], f1: ProximableFunctional, /) -> None: ...

    @overload
    def __init__(
        self: ProximableFunctionalSeparableSum[torch.Tensor, torch.Tensor],
        f1: ProximableFunctional,
        f2: ProximableFunctional,
        /,
    ) -> None: ...

    @overload
    def __init__(
        self: ProximableFunctionalSeparableSum[torch.Tensor, torch.Tensor, torch.Tensor],
        f1: ProximableFunctional,
        f2: ProximableFunctional,
        f3: ProximableFunctional,
        /,
    ) -> None: ...

    @overload
    def __init__(
        self: ProximableFunctionalSeparableSum[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor],
        f1: ProximableFunctional,
        f2: ProximableFunctional,
        f3: ProximableFunctional,
        f4: ProximableFunctional,
        /,
    ) -> None: ...

    @overload
    def __init__(
        self: ProximableFunctionalSeparableSum[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor],
        f1: ProximableFunctional,
        f2: ProximableFunctional,
        f3: ProximableFunctional,
        f4: ProximableFunctional,
        f5: ProximableFunctional,
        /,
    ) -> None: ...

    @overload
    def __init__(
        self: ProximableFunctionalSeparableSum[
            torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor, Unpack[tuple[torch.Tensor, ...]]
        ],
        f1: ProximableFunctional,
        f2: ProximableFunctional,
        f3: ProximableFunctional,
        f4: ProximableFunctional,
        f5: ProximableFunctional,
        /,
        *f: ProximableFunctional,
    ) -> None: ...

    def __init__(self, *functionals: ProximableFunctional) -> None:
        """Initialize the separable sum of proximable functionals.

        Parameters
        ----------
        functionals
            The proximable functionals to be summed.
        """
        super().__init__()
        self.functionals = functionals

    def __call__(self, *x: Unpack[T]) -> tuple[torch.Tensor]:
        """Evaluate the sum of separable functionals.

        Parameters
        ----------
        *x
            Input tensors. The number of input tensors must match the number
            of functionals in the sum.

        Returns
        -------
            Sum of the functionals applied to their respective inputs.
        """
        return super().__call__(*x)

    def forward(self, *x: Unpack[T]) -> tuple[torch.Tensor]:
        """Apply forward of ProximableFunctionalSeparableSum.

        .. note::
            Prefer calling the instance of the ProximableFunctionalSeparableSum operator as ``operator(x)`` over
            directly calling this method. See this PyTorch `discussion <https://discuss.pytorch.org/t/is-model-forward-x-the-same-as-model-call-x/33460/3>`_.
        """
        if len(x) != len(self.functionals):
            raise ValueError('The number of inputs must match the number of functionals.')
        result = reduce(
            operator.add, (f(xi)[0] for f, xi in zip(self.functionals, cast(tuple[torch.Tensor, ...], x), strict=True))
        )
        return (result,)

    def prox(self, *x: Unpack[T], sigma: float | torch.Tensor = 1) -> tuple[Unpack[T]]:
        """Apply the proximal operators of the functionals to the inputs.

        Parameters
        ----------
        x
            The inputs to the proximal operators
        sigma
            The scaling factor for the proximal operators

        Returns
        -------
            A tuple of the proximal operators applied to the inputs
        """
        prox_x = tuple(
            f.prox(xi, sigma)[0] for f, xi in zip(self.functionals, cast(tuple[torch.Tensor, ...], x), strict=True)
        )
        return cast(tuple[Unpack[T]], prox_x)

    def prox_convex_conj(self, *x: Unpack[T], sigma: float | torch.Tensor = 1) -> tuple[Unpack[T]]:
        """Apply the proximal operators of the convex conjugate of the functionals to the inputs.

        Parameters
        ----------
        x
            The inputs to the proximal operators
        sigma
            The scaling factor for the proximal operators

        Returns
        -------
            A tuple of the proximal convex conjugate operators applied to the inputs
        """
        prox_convex_conj_x = tuple(
            f.prox_convex_conj(xi, sigma)[0]
            for f, xi in zip(self.functionals, cast(tuple[torch.Tensor, ...], x), strict=True)
        )
        return cast(tuple[Unpack[T]], prox_convex_conj_x)

    @overload
    def __or__(
        self: ProximableFunctionalSeparableSum[Unpack[T]], other: ProximableFunctional
    ) -> ProximableFunctionalSeparableSum[Unpack[T], torch.Tensor]: ...

    @overload
    def __or__(
        self: ProximableFunctionalSeparableSum[Unpack[T]], other: ProximableFunctionalSeparableSum[torch.Tensor]
    ) -> ProximableFunctionalSeparableSum[Unpack[T], torch.Tensor]: ...

    @overload
    def __or__(
        self: ProximableFunctionalSeparableSum[Unpack[T]],
        other: ProximableFunctionalSeparableSum[torch.Tensor, torch.Tensor],
    ) -> ProximableFunctionalSeparableSum[Unpack[T], torch.Tensor, torch.Tensor]: ...

    @overload
    def __or__(
        self: ProximableFunctionalSeparableSum[Unpack[T]],
        other: ProximableFunctionalSeparableSum[torch.Tensor, torch.Tensor, torch.Tensor],
    ) -> ProximableFunctionalSeparableSum[Unpack[T], torch.Tensor, torch.Tensor, torch.Tensor]: ...

    @overload
    def __or__(
        self: ProximableFunctionalSeparableSum[Unpack[T]],
        other: ProximableFunctionalSeparableSum[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor],
    ) -> ProximableFunctionalSeparableSum[Unpack[T], torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]: ...

    @overload
    def __or__(
        self: ProximableFunctionalSeparableSum[Unpack[T]],
        other: ProximableFunctionalSeparableSum[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor],
    ) -> ProximableFunctionalSeparableSum[
        Unpack[T], torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor
    ]: ...

    def __or__(
        self: ProximableFunctionalSeparableSum,
        other: ProximableFunctional | ProximableFunctionalSeparableSum,
    ) -> ProximableFunctionalSeparableSum:
        """Separable sum of functionals.

        ``f | g`` is a ~mrpro.operators.ProximableFunctionalSeparableSum,
        with ``(f|g)(x,y) == f(x) + g(y)``.
        """
        if isinstance(other, ProximableFunctionalSeparableSum):
            return self.__class__(*self.functionals, *other.functionals)
        elif isinstance(other, ProximableFunctional):
            return self.__class__(*self.functionals, other)
        else:
            return NotImplemented

    def __ror__(
        self: ProximableFunctionalSeparableSum[Unpack[T]], other: ProximableFunctional
    ) -> ProximableFunctionalSeparableSum[torch.Tensor, Unpack[T]]:
        """Separable sum of functionals."""
        if isinstance(other, ProximableFunctional):
            return cast(
                ProximableFunctionalSeparableSum[torch.Tensor, Unpack[T]], self.__class__(other, *self.functionals)
            )
        else:
            return NotImplemented

    def __iter__(self) -> Iterator[ProximableFunctional]:
        """Iterate over the functionals."""
        return iter(self.functionals)

    def __len__(self) -> int:
        """Return the number of functionals."""
        return len(self.functionals)



"""K-space trajectory base class."""

from abc import ABC, abstractmethod

import torch
from einops import repeat

from mrpro.data.KTrajectory import KTrajectory
from mrpro.data.SpatialDimension import SpatialDimension


class KTrajectoryCalculator(ABC):
    """Base class for k-space trajectories."""

    @abstractmethod
    def __call__(
        self,
        *,
        n_k0: int,
        k0_center: int | torch.Tensor,
        k1_idx: torch.Tensor,
        k1_center: int | torch.Tensor,
        k2_idx: torch.Tensor,
        k2_center: int | torch.Tensor,
        encoding_matrix: SpatialDimension,
        reversed_readout_mask: torch.Tensor | None = None,
    ) -> KTrajectory:
        """Calculate the trajectory for given KHeader.

        The shapes of kz, ky and kx of the calculated trajectory must be
        broadcastable to (prod(all_other_dimensions), k2, k1, k0).

        Not all of the parameters will be used by all implementations.

        Parameters
        ----------
        n_k0
            number of samples in k0
        k0_center
            position of k-space center in k0
        k1_idx
            indices of k1
        k1_center
            position of k-space center in k1
        k2_idx
            indices of k2
        k2_center
            position of k-space center in k2
        reversed_readout_mask
            boolean tensor indicating reversed readout
        encoding_matrix
            encoding matrix

        Returns
        -------
            Trajectory

        """

    def _readout(
        self, n_k0: int, k0_center: int | torch.Tensor, reversed_readout_mask: torch.Tensor | None
    ) -> torch.Tensor:
        """Calculate the trajectory along one readout (k0 dimension).

        Parameters
        ----------
        n_k0
            number of samples in readout
        k0_center
            index of k-space center in readout
        reversed_readout_mask
            boolean tensor indicating reversed readout, e.g bipolar readout

        Returns
        -------
            trajectory along one readout

        """
        if isinstance(k0_center, int):
            k0_center = repeat(
                torch.tensor(k0_center), '... -> ... other coils k2 k1 k0', other=1, coils=1, k2=1, k1=1, k0=1
            )
        elif k0_center.ndim < 5:
            raise ValueError(f'Expected k0_center to have at least 5 dimensions, got {k0_center.ndim}.')
        k0 = torch.arange(n_k0, dtype=torch.float32) - k0_center
        if reversed_readout_mask is not None:
            # The -1 should not be required as far as we understand ISMRMRD, but
            # at least our Siemens epi data needs it.
            # See https://github.com/PTB-MR/mrpro/pull/832 for more details.
            k0 = torch.where(reversed_readout_mask.unsqueeze(-1), -1 - k0, k0)
        return k0

    def __repr__(self) -> str:
        """Get string representation."""
        return f'{type(self).__name__}()'


class DummyTrajectory(KTrajectoryCalculator):
    """Simple Dummy trajectory that returns zeros.

    Shape will fit to all data. Only used as dummy for testing.
    """

    def __call__(self, n_k0: int, k1_idx: torch.Tensor, k2_idx: torch.Tensor, **_) -> KTrajectory:
        """Calculate dummy trajectory."""
        shape = torch.broadcast_shapes(k1_idx.shape, k2_idx.shape)
        kx = torch.arange(shape.numel()).reshape(shape)
        ky = torch.zeros(*shape)
        kz = torch.arange(n_k0).reshape(1, 1, 1, n_k0)
        return KTrajectory(kz, ky, kx)



"""Base class for all data classes."""

import dataclasses
from collections.abc import Callable, Iterator, Sequence
from copy import copy as shallowcopy
from copy import deepcopy
from typing import ClassVar, TypeAlias, cast

import einops
import torch
from typing_extensions import Any, Protocol, Self, TypeVar, dataclass_transform, overload, runtime_checkable

from mrpro.utils.indexing import HasIndex, Indexer
from mrpro.utils.reduce_repeat import reduce_repeat
from mrpro.utils.reshape import broadcasted_concatenate, broadcasted_rearrange
from mrpro.utils.summarize import summarize_object
from mrpro.utils.typing import TorchIndexerType


@runtime_checkable
class HasReduceRepeats(Protocol):
    """Objects that have a _reduce_repeats method."""

    def _reduce_repeats_(self, tol: float = 1e-6, dim: Sequence[int] | None = None, recurse: bool = True) -> Self: ...


@runtime_checkable
class HasBroadcastedRearrange(Protocol):
    """Objects that have a _broadcasted_rearrange method."""

    def _broadcasted_rearrange(
        self, pattern: str, broadcasted_shape: Sequence[int], reduce_views: bool = True, **axes_lengths
    ) -> Self: ...


@runtime_checkable
class HasConcatenate(Protocol):
    """Objects that have a concatenate method."""

    def concatenate(self, *others: Self, dim: int) -> Self:
        """Concatenate other instances to self."""


class InconsistentDeviceError(RuntimeError):
    """Raised if the devices of different fields differ.

    There is no single device that all fields are on, thus
    the overall device of the object cannot be determined.
    """

    def __init__(self, *devices):
        """Initialize.

        Parameters
        ----------
        devices
            The devices of the fields that differ.
        """
        super().__init__(f'Inconsistent devices found, found at least {", ".join(str(d) for d in devices)}')


class InconsistentShapeError(RuntimeError):
    """Raised if fields are not broadcastable.

    The fields cannot be broadcasted to a common shape.
    """

    def __init__(self, *shapes):
        """Initialize.

        Parameters
        ----------
        shapes
            The shapes of the fields.
        """
        super().__init__(f'The shapes of the fields are not broadcastable. Found shapes: {shapes}.')


T = TypeVar('T')


@dataclass_transform()
class Dataclass:
    """A supercharged dataclass with additional functionality.

    This class extends the functionality of the standard `dataclasses.dataclass` by adding:

    - a `apply` method to apply a function to all fields
    - a `~Dataclass.clone` method to create a deep copy of the object
    - `~Dataclass.to`, `~Dataclass.cpu`, `~Dataclass.cuda` methods to move all tensor fields to a device.

    It is intended to be used as a base class for all dataclasses in the `mrpro` package.
    """

    __dataclass_fields__: ClassVar[dict[str, dataclasses.Field[Any]]]
    __auto_reduce_repeats: bool
    __initialized: bool

    def __init_subclass__(  # noqa: D417
        cls,
        no_new_attributes: bool = True,
        auto_reduce_repeats: bool = True,
        init: bool = True,
        *args,
        **kwargs,
    ) -> None:
        """Create a new dataclass subclass.

        Parameters
        ----------
        no_new_attributes
            If `True`, new attributes cannot be added to the class after it is created.
        auto_reduce_repeats
            If `True`, try to reduce dimensions only containing repeats to singleton.
            This will be done after init and post_init.
        init
            If `True`, an automatic init function will be added. Set to `False` to use a custom init.
        """
        dataclasses.dataclass(cls, repr=False, eq=False, init=init)
        super().__init_subclass__(**kwargs)
        child_post_init = vars(cls).get('__post_init__')

        if no_new_attributes:

            def new_setattr(self: Dataclass, name: str, value: Any) -> None:  # noqa: ANN401
                """Set an attribute."""
                if not hasattr(self, name) and hasattr(self, '_Dataclass__initialized'):
                    raise AttributeError(f'Cannot set attribute {name} on {self.__class__.__name__}')
                object.__setattr__(self, name, value)

            cls.__setattr__ = new_setattr  # type: ignore[method-assign, assignment]

        cls.__auto_reduce_repeats = auto_reduce_repeats

        if child_post_init and child_post_init is not Dataclass.__post_init__:

            def chained_post_init(self: Dataclass, *args, **kwargs) -> None:
                child_post_init(self, *args, **kwargs)
                Dataclass.__post_init__(self)

            cls.__post_init__ = chained_post_init  # type: ignore[method-assign]

    def __post_init__(self) -> None:
        """Can be overridden in subclasses to add custom initialization logic."""
        self.__initialized = True
        if self.__auto_reduce_repeats:
            self._reduce_repeats_(recurse=False)

    def _reduce_repeats_(self, tol: float = 1e-6, dim: Sequence[int] | None = None, recurse: bool = True) -> Self:
        """Reduce repeated dimensions in fields to singleton.

        Parameters
        ----------
        tol
            tolerance.
        dim
            dimensions to try to reduce to singletons. `None` means all.
        recurse
            recurse into dataclass fields.
        """

        def apply_reduce(data: T) -> T:
            if isinstance(data, torch.Tensor):
                return cast(T, reduce_repeat(data, tol, dim))
            if isinstance(data, HasReduceRepeats) and not isinstance(data, Dataclass):
                data._reduce_repeats_(tol, dim)
            return cast(T, data)

        return self.apply_(apply_reduce, recurse=recurse)

    def items(self) -> Iterator[tuple[str, Any]]:
        """Get an iterator over names and values of fields."""
        for field in dataclasses.fields(self):
            name = field.name
            data = getattr(self, name)
            yield name, data

    def apply_(
        self: Self,
        function: Callable[[Any], Any] | None = None,
        *,
        memo: dict[int, Any] | None = None,
        recurse: bool = True,
    ) -> Self:
        """Apply a function to all children in-place.

        Parameters
        ----------
        function
            The function to apply to all fields. `None` is interpreted as a no-op.
        memo
            A dictionary to keep track of objects that the function has already been applied to,
            to avoid multiple applications. This is useful if the object has a circular reference.
        recurse
            If `True`, the function will be applied to all children that are `Dataclass` instances.
        """
        applied: Any

        if memo is None:
            memo = {}

        if function is None:
            return self

        for name, data in self.items():
            if id(data) in memo:
                # this works even if self is frozen
                object.__setattr__(self, name, memo[id(data)])
                continue
            if recurse and isinstance(data, Dataclass):
                applied = data.apply_(function, memo=memo)
            else:
                applied = function(data)
            memo[id(data)] = applied
            object.__setattr__(self, name, applied)
        return self

    def apply(
        self: Self,
        function: Callable[[Any], Any] | None = None,
        *,
        recurse: bool = True,
    ) -> Self:
        """Apply a function to all children. Returns a new object.

        Parameters
        ----------
        function
            The function to apply to all fields. `None` is interpreted as a no-op.
        recurse
            If `True`, the function will be applied to all children that are `Dataclass` instances.
        """
        new = self.clone().apply_(function, recurse=recurse)
        return new

    # region Move to device/dtype
    @overload
    def to(
        self,
        device: str | torch.device | int | None = None,
        dtype: torch.dtype | None = None,
        non_blocking: bool = False,
        *,
        copy: bool = False,
        memory_format: torch.memory_format | None = None,
    ) -> Self: ...

    @overload
    def to(
        self,
        dtype: torch.dtype,
        non_blocking: bool = False,
        *,
        copy: bool = False,
        memory_format: torch.memory_format | None = None,
    ) -> Self: ...

    @overload
    def to(
        self,
        tensor: torch.Tensor,
        non_blocking: bool = False,
        *,
        copy: bool = False,
        memory_format: torch.memory_format | None = None,
    ) -> Self: ...

    def to(self, *args, **kwargs) -> Self:
        """Perform dtype and/or device conversion of data.

        A `torch.dtype` and `torch.device` are inferred from the arguments
        args and kwargs. Please have a look at the
        documentation of `torch.Tensor.to` for more details.

        A new instance of the dataclass will be returned.

        The conversion will be applied to all Tensor- or Module
        fields of the dataclass, and to all fields that implement
        the `Dataclass`.

        The dtype-type, i.e. float or complex will always be preserved,
        but the precision of floating point dtypes might be changed.

        Examples
        --------
        If called with ``dtype=torch.float32`` OR ``dtype=torch.complex64``:

        - A ``complex128`` tensor will be converted to ``complex64``
        - A ``float64`` tensor will be converted to ``float32``
        - A ``bool`` tensor will remain ``bool``
        - An ``int64`` tensor will remain ``int64``

        If other conversions are desired, please use the `~torch.Tensor.to` method of
        the fields directly.

        If the copy argument is set to `True` (default), a deep copy will be returned
        even if no conversion is necessary.
        If two fields are views of the same data before, in the result they will be independent
        copies if copy is set to `True` or a conversion is necessary.
        If set to `False`, some Tensors might be shared between the original and the new object.
        """
        # Parse the arguments of the three overloads and call _to with the parsed arguments
        parsedType: TypeAlias = tuple[
            str | torch.device | int | None, torch.dtype | None, bool, bool, torch.memory_format
        ]

        def parse3(
            other: torch.Tensor,
            non_blocking: bool = False,
            copy: bool = False,
        ) -> parsedType:
            return other.device, other.dtype, non_blocking, copy, torch.preserve_format

        def parse2(
            dtype: torch.dtype,
            non_blocking: bool = False,
            copy: bool = False,
            memory_format: torch.memory_format = torch.preserve_format,
        ) -> parsedType:
            return None, dtype, non_blocking, copy, memory_format

        def parse1(
            device: str | torch.device | int | None = None,
            dtype: None | torch.dtype = None,
            non_blocking: bool = False,
            copy: bool = False,
            memory_format: torch.memory_format = torch.preserve_format,
        ) -> parsedType:
            return device, dtype, non_blocking, copy, memory_format

        if (args and isinstance(args[0], torch.Tensor)) or 'tensor' in kwargs:
            # overload 3 ("tensor" specifies the dtype and device)
            device, dtype, non_blocking, copy, memory_format = parse3(*args, **kwargs)
        elif args and isinstance(args[0], torch.dtype):
            # overload 2 (no device specified, only dtype)
            device, dtype, non_blocking, copy, memory_format = parse2(*args, **kwargs)
        else:
            # overload 1 (device and dtype specified)
            device, dtype, non_blocking, copy, memory_format = parse1(*args, **kwargs)
        return self._to(device=device, dtype=dtype, non_blocking=non_blocking, memory_format=memory_format, copy=copy)

    def _to(
        self,
        device: torch.device | str | int | None = None,
        dtype: torch.dtype | None = None,
        non_blocking: bool = False,
        memory_format: torch.memory_format = torch.preserve_format,
        shared_memory: bool = False,
        copy: bool = False,
        memo: dict | None = None,
    ) -> Self:
        """Move data to device and convert dtype if necessary.

        This method is called by `.to()`, `.cuda()`, `.cpu()`,
        `.double()`, and so on. It should not be called directly.

        See `Dataclass.to()` for more details.

        Parameters
        ----------
        device
            The destination device.
        dtype
            The destination dtype.
        non_blocking
            If `True` and the source is in pinned memory, the copy will be asynchronous with respect to the host.
            Otherwise, the argument has no effect.
        memory_format
            The desired memory format of returned tensor.
        shared_memory
            If `True` and the target device is CPU, the tensors will reside in shared memory.
            Otherwise, the argument has no effect.
        copy
            If `True`, the returned tensor will always be a copy, even if the input was already on the correct device.
            This will also create new tensors for views.
        memo
            A dictionary to keep track of already converted objects to avoid multiple conversions.
        """
        new = shallowcopy(self) if copy or not isinstance(self, torch.nn.Module) else self

        if memo is None:
            memo = {}

        def _tensor_to(data: torch.Tensor) -> torch.Tensor:
            """Move tensor to device and convert dtype if necessary."""
            new_dtype: torch.dtype | None
            if dtype is not None and data.dtype.is_floating_point:
                new_dtype = dtype.to_real()
            elif dtype is not None and data.dtype.is_complex:
                new_dtype = dtype.to_complex()
            else:
                # bool or int: keep as is
                new_dtype = None
            data = data.to(
                device,
                new_dtype,
                non_blocking=non_blocking,
                memory_format=memory_format,
                copy=copy,
            )
            if shared_memory:
                data.share_memory_()
            return data

        def _module_to(data: torch.nn.Module) -> torch.nn.Module:
            if copy:
                data = deepcopy(data)
            return data._apply(_tensor_to, recurse=True)

        def _mixin_to(obj: Dataclass) -> Dataclass:
            return obj._to(
                device=device,
                dtype=dtype,
                non_blocking=non_blocking,
                memory_format=memory_format,
                shared_memory=shared_memory,
                copy=copy,
                memo=memo,
            )

        def _convert(data: T) -> T:
            converted: Any  # https://github.com/python/mypy/issues/10817
            if isinstance(data, torch.Tensor):
                converted = _tensor_to(data)
            elif isinstance(data, Dataclass):
                converted = _mixin_to(data)
            elif isinstance(data, torch.nn.Module):
                converted = _module_to(data)
            else:
                converted = data
            return cast(T, converted)

        # manual recursion allows us to do the copy only once
        new.apply_(_convert, memo=memo, recurse=False)
        return new

    def cuda(
        self,
        device: torch.device | str | int | None = None,
        *,
        non_blocking: bool = False,
        memory_format: torch.memory_format = torch.preserve_format,
        copy: bool = False,
    ) -> Self:
        """Put object in CUDA memory.

        Parameters
        ----------
        device
            The destination GPU device. Defaults to the current CUDA device.
        non_blocking
            If `True` and the source is in pinned memory, the copy will be asynchronous with respect to the host.
            Otherwise, the argument has no effect.
        memory_format
            The desired memory format of returned tensor.
        copy:
            If `True`, the returned tensor will always be a copy, even if the input was already on the correct device.
            This will also create new tensors for views.
        """
        if device is None:
            device = torch.device(torch.cuda.current_device())
        return self._to(device=device, dtype=None, memory_format=memory_format, non_blocking=non_blocking, copy=copy)

    def cpu(self, *, memory_format: torch.memory_format = torch.preserve_format, copy: bool = False) -> Self:
        """Put in CPU memory.

        Parameters
        ----------
        memory_format
            The desired memory format of returned tensor.
        copy
            If `True`, the returned tensor will always be a copy, even if the input was already on the correct device.
            This will also create new tensors for views.
        """
        return self._to(device='cpu', dtype=None, non_blocking=True, memory_format=memory_format, copy=copy)

    def double(self, *, memory_format: torch.memory_format = torch.preserve_format, copy: bool = False) -> Self:
        """Convert all float tensors to double precision.

        converts ``float`` to ``float64`` and ``complex`` to ``complex128``


        Parameters
        ----------
        memory_format
            The desired memory format of returned tensor.
        copy
            If `True`, the returned tensor will always be a copy, even if the input was already on the correct device.
            This will also create new tensors for views.
        """
        return self._to(dtype=torch.float64, memory_format=memory_format, copy=copy)

    def half(self, *, memory_format: torch.memory_format = torch.preserve_format, copy: bool = False) -> Self:
        """Convert all float tensors to half precision.

        converts ``float`` to ``float16`` and ``complex`` to ``complex32``


        Parameters
        ----------
        memory_format
            The desired memory format of returned tensor.
        copy
            If `True`, the returned tensor will always be a copy, even if the input was already on the correct device.
            This will also create new tensors for views.
        """
        return self._to(dtype=torch.float16, memory_format=memory_format, copy=copy)

    def single(self, *, memory_format: torch.memory_format = torch.preserve_format, copy: bool = False) -> Self:
        """Convert all float tensors to single precision.

        converts ``float`` to ``float32`` and ``complex`` to ``complex64``


        Parameters
        ----------
        memory_format
            The desired memory format of returned tensor.
        copy
            If `True`, the returned tensor will always be a copy, even if the input was already on the correct device.
            This will also create new tensors for views.
        """
        return self._to(dtype=torch.float32, memory_format=memory_format, copy=copy)

    # endregion Move to device/dtype

    def detach(self) -> Self:
        """Detach the data from the autograd graph.

        Returns
        -------
            A new dataclass with the data detached from the autograd graph.
            The data is shared between the original and the new object.
            Use ``detach().clone()`` to create an independent copy.
        """
        new = shallowcopy(self)
        new.apply_(lambda data: data.detach() if isinstance(data, torch.Tensor) else data, recurse=True)
        return new

    # region Properties
    @property
    def device(self) -> torch.device | None:
        """Return the device of the tensors.

        Looks at each field of a dataclass implementing a device attribute,
        such as `torch.Tensor` or `Dataclass` instances. If the devices
        of the fields differ, an :py:exc:`~mrpro.data.InconsistentDeviceError` is raised, otherwise
        the device is returned. If no field implements a device attribute,
        None is returned.

        Raises
        ------
        :py:exc:`InconsistentDeviceError`
            If the devices of different fields differ.

        Returns
        -------
            The device of the fields or `None` if no field implements a `device` attribute.
        """
        device: None | torch.device = None
        for _, data in self.items():
            if not hasattr(data, 'device'):
                continue
            current_device = getattr(data, 'device', None)
            if current_device is None:
                continue
            if device is None:
                device = current_device
            elif device != current_device:
                raise InconsistentDeviceError(current_device, device)
        return device

    def clone(self: Self) -> Self:
        """Return a deep copy of the object."""
        return self._to(device=None, dtype=None, non_blocking=False, memory_format=torch.preserve_format, copy=True)

    @property
    def is_cuda(self) -> bool:
        """Return `True` if all tensors are on a single CUDA device.

        Checks all tensor attributes of the dataclass for their device,
        (recursively if an attribute is a `Dataclass`)


        Returns `False` if not all tensors are on the same CUDA devices, or if the device is inconsistent,
        returns `True` if the data class has no tensors as attributes.
        """
        try:
            device = self.device
        except InconsistentDeviceError:
            return False
        if device is None:
            return True
        return device.type == 'cuda'

    @property
    def is_cpu(self) -> bool:
        """Return True if all tensors are on the CPU.

        Checks all tensor attributes of the dataclass for their device,
        (recursively if an attribute is a `Dataclass`)

        Returns `False` if not all tensors are on cpu or if the device is inconsistent,
        returns `True` if the data class has no tensors as attributes.
        """
        try:
            device = self.device
        except InconsistentDeviceError:
            return False
        if device is None:
            return True
        return device.type == 'cpu'

    @property
    def shape(self) -> torch.Size:
        """Return the broadcasted shape of all tensor/data fields.

        Each field of this dataclass is broadcastable to this shape.

        Returns
        -------
            The broadcasted shape of all fields.

        Raises
        ------
        :py:exc:`InconsistentShapeError`
            If the shapes cannot be broadcasted.
        """
        shapes = []
        for _, data in self.items():
            if not hasattr(data, 'shape'):
                continue
            shapes.append(data.shape)
        try:
            return torch.broadcast_shapes(*shapes)
        except RuntimeError:
            raise InconsistentShapeError(*shapes) from None

    @property
    def ndim(self) -> int:
        """Return the number of dimensions of the dataclass.

        This is the number of dimensions of the broadcasted shape of all fields.
        """
        return len(self.shape)

    # endregion Properties

    # region Representation
    def __repr__(self) -> str:
        """Get string representation of Dataclass."""
        header = [type(self).__name__]

        try:
            device = self.device
            if device:
                header.append(f'on device "{device}"')
        except RuntimeError:
            header.append('on mixed devices')

        try:
            if shape := self.shape:
                header.append(f'with (broadcasted) shape {list(shape)!s}')
        except RuntimeError:
            header.append('with inconsistent shape')

        output = ' '.join(header) + '.\n'

        output += '\n'.join(
            f'  {field.name}: {summarize_object(value)}'
            for field in dataclasses.fields(self)
            if not (field.name.startswith('_') or (value := getattr(self, field.name, None)) is None)
        )
        return output

    # We return the same for __repr__ and __str__.
    # This break the "_str_ if for users, _repr_ for developers rule" of python.
    # But it makes interactive work on repl or notebooks easier, as `obj` can be used instead
    # of `print(obj)`. It would be infeasable for most dataclasses to implement a proper  __repr__
    # that uniquely describes the data and could be used to recreate the object anyways.

    def __str__(self) -> str:
        """Return the same as __repr__."""
        return repr(self)

    def __shortstr__(self) -> str:
        """Return a short string representation."""
        output = type(self).__name__
        if self.shape:
            output = output + f'<{", ".join(map(str, self.shape))}>'
        return output

    # endregion Representation

    # region Indexing
    def __getitem__(self, index: TorchIndexerType | Indexer) -> Self:
        """Index the dataclass."""
        indexer = index if isinstance(index, Indexer) else Indexer(self.shape, index)
        memo: dict[int, Any] = {}

        def apply_index(data: T) -> T:
            if isinstance(data, torch.Tensor):
                return cast(T, indexer(data))
            if isinstance(data, Dataclass):
                indexed = shallowcopy(data)
                indexed.apply_(apply_index, memo=memo, recurse=False)
                return cast(T, indexed)
            if isinstance(data, HasIndex):
                # Rotation
                return cast(T, data._index(indexer))
            return cast(T, data)

        new = shallowcopy(self)
        new.apply_(apply_index, memo=memo, recurse=False)
        return new

    # endregion Indexing

    def rearrange(self, pattern: str, **axes_lengths: int) -> Self:
        """Rearrange the data according to the specified pattern.

        Similar to `einops.rearrange`, allowing flexible rearrangement of data dimensions.

        Examples
        --------
        >>> # Split the phase encode lines into 8 cardiac phases
        >>> data.rearrange('batch coils k2 (phase k1) k0 -> batch phase coils k2 k1 k0', phase=8)
        >>> # Split the k-space samples into 64 k1 and 64 k2 lines
        >>> data.rearrange('... 1 1 (k2 k1 k0) -> ... k2 k1 k0', k2=64, k1=64, k0=128)

        Parameters
        ----------
        pattern
            String describing the rearrangement pattern. See `einops.rearrange` and the examples above for more details.
        **axes_lengths : dict
            Optional dictionary mapping axis names to their lengths.
            Used when pattern contains unknown dimensions.

        Returns
        -------
            The rearranged data with the same type as the input.

        """
        memo: dict = {}
        shape = self.shape

        def apply_rearrange(data: T) -> T:
            def rearrange_tensor(data: torch.Tensor) -> torch.Tensor:
                return broadcasted_rearrange(data, pattern, shape, reduce_views=True, **axes_lengths)

            if isinstance(data, torch.Tensor):
                return cast(T, rearrange_tensor(data))
            if isinstance(data, HasBroadcastedRearrange):
                return cast(T, data._broadcasted_rearrange(pattern, shape, reduce_views=True, **axes_lengths))
            elif isinstance(data, Dataclass):
                return cast(T, shallowcopy(data).apply_(apply_rearrange, memo=memo, recurse=False))
            else:
                return data

        new = shallowcopy(self)
        return new.apply_(apply_rearrange, memo=memo, recurse=False)

    def swapdims(self, dim0: int, dim1: int) -> Self:
        """Swap two dimensions of the dataclass.

        Parameters
        ----------
        dim0
            First dimension to swap.
        dim1
            Second dimension to swap.

        Returns
        -------
            The dataclass with the dimensions swapped.
        """
        axes = [f'dim{i}' for i in range(self.ndim)]
        input_pattern = ' '.join(axes)
        axes[dim0], axes[dim1] = axes[dim1], axes[dim0]
        output_pattern = ' '.join(axes)
        return self.rearrange(f'{input_pattern} -> {output_pattern}')

    def split(self, dim: int, size: int = 1, overlap: int = 0, dilation: int = 1) -> tuple[Self, ...]:
        """Split the dataclass along a dimension.

        Parameters
        ----------
        dim
            dimension to split along.
        size
            size of the splits.
        overlap
            overlap between splits.
            The stride will be `size - overlap`.
            Negative overlap will leave spaces between splits.
        dilation
            dilation of elements in each split.

        Examples
        --------
        If the dimension has 6 elements:

        - split with size 2, overlap 0, dilation 1 -> elements (0,1), (2,3), and (4,5)
        - split with size 2, overlap 1, dilation 1 -> elements (0,1), (1,2), (2,3), (3,4), (4,5), and (5,6)
        - split with size 2, overlap 0, dilation 2 -> elements (0,2), and (3,5)
        - split with size 2, overlap -1, dilation 1 -> elements (0,1), and (3,4)


        Returns
        -------
            A tuple of the splits.
        """
        shape = self.shape
        if not -len(shape) <= dim < len(shape):
            raise ValueError(f'Dimension {dim} out of bounds for shape {shape}')
        if dilation < 1:
            raise ValueError('Dilation must be larger than 0')
        if overlap > size:
            raise ValueError('Overlap must be smaller than size')
        dim = dim % len(shape)
        indices = [
            (*[slice(None)] * dim, slice(start, start + size * dilation, dilation))
            for start in range(0, shape[dim] - size * dilation + 1 + max(0, overlap), size - overlap)
        ]
        res = tuple(self[idx] for idx in indices)
        return res

    def concatenate(self, *others: Self, dim: int) -> Self:
        """Concatenate other instances to the current instance.

        Only tensor-like fields will be concatenated in the specified dimension.
        List fields will be concatenated as a list.
        Other fields will be ignored.

        Parameters
        ----------
        others
            other instance to concatnate.
        dim
            The dimension to concatenate along.

        Returns
        -------
            The concatenated dataclass.
        """
        new = shallowcopy(self)
        shapes = [self.shape, *[other.shape for other in others]]
        for field in dataclasses.fields(new):
            value_self = getattr(new, field.name)
            value_others = [getattr(other, field.name) for other in others]
            if all(isinstance(v, list) for v in (value_self, *value_others)):
                for v in value_others:
                    value_self.extend(v)
            elif all(isinstance(v, torch.Tensor) for v in (value_self, *value_others)):
                tensors = [t.broadcast_to(s) for t, s in zip((value_self, *value_others), shapes, strict=True)]
                setattr(new, field.name, broadcasted_concatenate(tensors, dim=dim))
            elif isinstance(value_self, HasConcatenate):
                setattr(new, field.name, value_self.concatenate(*value_others, dim=dim))
        new._reduce_repeats_(recurse=True)
        return new

    def stack(self, *others: Self) -> Self:
        """Stack other along new first dimension.

        Parameters
        ----------
        others
            other instance to stack.
        """
        return self[None].concatenate(*[o[None] for o in others], dim=0)

    def __eq__(self, other: object) -> bool:
        """Check deep equality of two dataclasses.

        Tests equality up to broadcasting.
        """
        if not isinstance(other, type(self)):
            return False
        if self is other:
            return True
        for field in dataclasses.fields(self):
            field_self = getattr(self, field.name)
            field_other = getattr(other, field.name)
            if not isinstance(field_self, type(field_other)):
                return False
            elif isinstance(field_self, torch.Tensor):
                try:
                    if not torch.equal(*torch.broadcast_tensors(field_self, field_other)):
                        return False
                except RuntimeError:
                    return False
            elif field_self != field_other:
                return False
        return True

    def __len__(self) -> int:
        """Return the number of fields in the dataclass along the first dimension."""
        return self.shape[0]

    def __iter__(self) -> Iterator[Any]:
        """Iterate over the first dimension of the dataclass."""
        for i in range(len(self)):
            yield self[i]


class FakeDataclassBackend(einops._backends.AbstractBackend):
    """Einops backend for Dataclass: Will only raise an error if used."""

    framework_name = 'mrpro.data.Dataclass'

    def is_appropriate_type(self, x) -> bool:  # noqa: ANN001
        """Check if the object is a Dataclass."""
        if isinstance(x, Dataclass):
            raise NotImplementedError(
                'To use einops with mrpro dataclasses, please use the rearrange method of an dataclass instance.'
            )
        return False



"""Exponential Moving Average (EMA) dictionary."""

from collections.abc import ItemsView, KeysView, Mapping, ValuesView
from typing import Any

import torch


class EMADict:
    """
    Exponential Moving Average (EMA) dictionary.

    Maintains an EMA of values for each key. On update, existing keys are
    updated with EMA, and new keys are added directly.

    Detaches the values from the autograd graph.


    """

    def __init__(
        self,
        decay: float,
    ):
        """Initialize the EMA dictionary.

        Parameters
        ----------
        decay
            Decay rate for EMA (between 0 and 1).
        """
        self.decay: float = decay
        if not 0 <= decay <= 1:
            raise ValueError(f'Decay must be between 0 and 1, got {decay}')
        self._data: dict[str, Any] = {}

    def __getitem__(self, key: str) -> Any:  # noqa: ANN401
        """Get the value of the EMA dict for a given key."""
        return self._data[key]

    def __setitem__(self, key: str, value: Any) -> None:  # noqa: ANN401
        """Set the value of the EMA dict for a given key."""
        if key in self._data:
            old_v = self._data[key]
            if isinstance(old_v, torch.Tensor) and isinstance(value, torch.Tensor):
                if torch.is_floating_point(old_v) or torch.is_complex(old_v):
                    old_v.mul_(self.decay).add_(value.detach().to(old_v.device), alpha=1.0 - self.decay)
                else:
                    old_v.copy_(value)
                return
            elif isinstance(old_v, float) and isinstance(value, float):  # noqa: SIM114
                self._data[key] = self.decay * old_v + (1.0 - self.decay) * value
                return
            elif isinstance(old_v, complex) and isinstance(value, complex):
                self._data[key] = self.decay * old_v + (1.0 - self.decay) * value
                return

        if isinstance(value, torch.Tensor):
            self._data[key] = value.detach().clone()
            return

        self._data[key] = value

    def __delitem__(self, key: str) -> None:
        """Delete a key from the EMA dict."""
        del self._data[key]

    def __contains__(self, key: str) -> bool:
        """Check if a key is in the EMA dict."""
        return key in self._data

    def values(self) -> ValuesView[Any]:
        """Get the values of the EMA dict."""
        return self._data.values()

    def keys(self) -> KeysView[str]:
        """Get the keys of the EMA dict."""
        return self._data.keys()

    def items(self) -> ItemsView[str, Any]:
        """Get the items of the EMA dict."""
        return self._data.items()

    def update(self, other: Mapping[Any, Any]) -> None:
        """Update the EMA dict with another dictionary.

        For existing keys, performs EMA update. For new keys, adds them directly.

        Parameters
        ----------
        other
            Dictionary to update from.
        """
        for k, v in other.items():
            self.__setitem__(k, v)



"""Identity Operator with arbitrary number of inputs."""

import torch
from typing_extensions import Self

from mrpro.operators.EndomorphOperator import EndomorphOperator, endomorph


class MultiIdentityOp(EndomorphOperator):
    r"""The Identity Operator.

    An endomorph Operator that returns multiple inputs unchanged.
    """

    @endomorph
    def __call__(self, *x: torch.Tensor) -> tuple[torch.Tensor, ...]:
        """Apply the multi-input identity operation.

        This operator returns all input tensors unchanged.

        Parameters
        ----------
        *x
            One or more input tensors.

        Returns
        -------
            The input tensors, unchanged.
        """
        return super().__call__(*x)

    @endomorph
    def forward(self, *x: torch.Tensor) -> tuple[torch.Tensor, ...]:
        """Apply forward of MultiIdentityOp.

        .. note::
            Prefer calling the instance of the MultiIdentityOp operator as ``operator(x)`` over
            directly calling this method. See this PyTorch `discussion <https://discuss.pytorch.org/t/is-model-forward-x-the-same-as-model-call-x/33460/3>`_.
        """
        return x

    @endomorph
    def adjoint(self, *x: torch.Tensor) -> tuple[torch.Tensor, ...]:
        """Apply the adjoint of the multi-input identity operation.

        Since the multi-input identity operator is self-adjoint, this method
        returns all input tensors unchanged.

        Parameters
        ----------
        *x
            One or more input tensors.

        Returns
        -------
            The input tensors, unchanged.
        """
        return x

    @property
    def H(self) -> Self:  # noqa: N802
        """Adjoint Identity."""
        return self



"""Non-Uniform Fast Fourier Operator."""

from collections.abc import Sequence
from functools import partial
from itertools import product
from types import EllipsisType
from typing import Literal

import torch
from pytorch_finufft.functional import finufft_type1, finufft_type2
from typing_extensions import Self

from mrpro.data.KTrajectory import KTrajectory
from mrpro.data.SpatialDimension import SpatialDimension
from mrpro.operators.FastFourierOp import FastFourierOp
from mrpro.operators.LinearOperator import LinearOperator


class NonUniformFastFourierOp(LinearOperator, adjoint_as_backward=True):
    """Non-Uniform Fast Fourier Operator class."""

    def __init__(
        self,
        direction: Sequence[Literal['x', 'y', 'z', -3, -2, -1]],
        recon_matrix: SpatialDimension[int] | Sequence[int],
        encoding_matrix: SpatialDimension[int] | Sequence[int],
        traj: KTrajectory,
        oversampling: float = 2.0,
    ) -> None:
        """Initialize Non-Uniform Fast Fourier Operator.

        .. note::
            Consider using `~mrpro.operators.FourierOp` instead of this operator. It automatically detects if a
            non-uniform or regular fast Fourier transformation is required and can also be constructed automatically
            from a `mrpro.data.KData` object.

        .. note::
            The NUFFT is scaled such that it matches 'orthonormal' FFT scaling for cartesian trajectories.
            This is different from other packages, which apply scaling based on the size of the oversampled grid.

        Parameters
        ----------
        direction
            direction along which non-uniform FFT is applied
        recon_matrix
            Dimension of the reconstructed image. If this is `~mrpro.data.SpatialDimension` only values of directions
            will be used. Otherwise, it should be a `Sequence` of the same length as direction.
        encoding_matrix
            Dimension of the encoded k-space. If this is `~mrpro.data.SpatialDimension` only values of directions will
            be used. Otherwise, it should be a `Sequence` of the same length as direction.
        traj
            The k-space trajectories where the frequencies are sampled.
        oversampling
            Oversampling used for interpolation in non-uniform FFTs.
            On GPU, 2.0 uses an optimized kernel, any value > 1.0 will work.
            On CPU, there are kernels for 2.0 and 1.25. The latter saves memory. Set to 0.0 for automatic selection.
        """
        super().__init__()

        # Convert to negative indexing
        direction_dict = {'z': -3, 'y': -2, 'x': -1, -3: -3, -2: -2, -1: -1}
        self._direction_zyx = tuple(direction_dict[d] for d in direction)
        if len(direction) != len(set(self._direction_zyx)):
            raise ValueError(f'Directions must be unique. Normalized directions are {self._direction_zyx}')
        if not self._direction_zyx:
            return
        if len(self._direction_zyx) != 1 and len(self._direction_zyx) != 2 and len(self._direction_zyx) != 3:
            raise ValueError('Only 0D, 1D, 2D or 3D NUFFT is supported')

        trajectory = [
            ks for ks, i in zip((traj.kz, traj.ky, traj.kx), (-3, -2, -1), strict=True) if i in self._direction_zyx
        ]

        # Find out along which dimensions (k0, k1 or k2) nufft needs to be applied, i.e. where it is not singleton
        dimension_210: list[int] = []
        for dim in (-3, -2, -1):
            for ks in trajectory:
                if ks.shape[dim] > 1:
                    dimension_210.append(dim)
                    break  # one case where nufft is needed is enough for each dimension

        # For e.g. single shot acquisitions the number of dimensions do not necessarily match the number of
        # directions. This leads to a mismatch between reconstructed and expected dimensions. To avoid this we try
        # to find the most logical solution, i.e. add another singleton direction
        if len(self._direction_zyx) > len(dimension_210):
            for dim in (-1, -2, -3):
                if dim not in dimension_210 and all(ks.shape[dim] == 1 for ks in (traj.kz, traj.ky, traj.kx)):
                    dimension_210.append(dim)
                if len(self._direction_zyx) == len(dimension_210):
                    break
            dimension_210.sort()
        self._dimension_210 = tuple(dimension_210)

        if len(self._direction_zyx) != len(self._dimension_210):
            raise ValueError(
                f'Mismatch between number of nufft directions {self._direction_zyx} and dims {dimension_210}'
            )

        if isinstance(recon_matrix, SpatialDimension):
            im_size = tuple([recon_matrix.zyx[d] for d in self._direction_zyx])
        else:
            if (n_recon_matrix := len(recon_matrix)) != (n_nufft_dir := len(self._direction_zyx)):
                raise ValueError(f'recon_matrix should have {n_nufft_dir} entries but has {n_recon_matrix}')
            im_size = tuple(recon_matrix)
        assert len(im_size) == 1 or len(im_size) == 2 or len(im_size) == 3  # mypy  # noqa: S101

        if isinstance(encoding_matrix, SpatialDimension):
            k_size = tuple([int(encoding_matrix.zyx[d]) for d in self._direction_zyx])
        else:
            if (n_enc_matrix := len(encoding_matrix)) != (n_nufft_dir := len(self._direction_zyx)):
                raise ValueError(f'encoding_matrix should have {n_nufft_dir} entries but has {n_enc_matrix}')
            k_size = tuple(encoding_matrix)

        omega_list = [
            k * 2 * torch.pi / ks
            for k, ks in zip(
                trajectory,
                k_size,
                strict=True,
            )
        ]
        # Broadcast shapes not always needed but also does not hurt
        omega_list = list(torch.broadcast_tensors(*omega_list))
        omega = torch.concatenate(omega_list, dim=-4)  # use the 'coil' dim for the direction
        self._traj_broadcast_shape = omega.shape
        keep_dims_210 = [-4, *dimension_210]  # -4 is always coil
        permute_210 = [i for i in range(-omega.ndim, 0) if i not in keep_dims_210] + keep_dims_210
        # omega should be (sep_dims, 1, 2 or 3, nufft_dimensions)
        omega = omega.permute(*permute_210)
        omega = omega.flatten(end_dim=-len(keep_dims_210) - 1).flatten(start_dim=-len(keep_dims_210) + 1)
        # scaling independent of nufft oversampling, matches FFT scaling for cartesian trajectories
        self.scale = torch.tensor(k_size).prod().sqrt().reciprocal()
        self.oversampling = oversampling

        # we want to rearrange everything into (sep_dims)(joint_dims)(nufft_dims) where sep_dims are dimension
        # where the traj changes, joint_dims are dimensions where the traj does not change and nufft_dims are the
        # dimensions along which the nufft is applied. We have to do this for the (z-y-x) and (k2-k1-k0) space
        # separately. If we know two of the three dimensions we can infer the rest. We cannot do the other
        # dimensions here because they might be different between data and trajectory.
        self._joint_dims_210 = [
            d for d in [-3, -2, -1] if d not in dimension_210 and self._traj_broadcast_shape[d] == 1
        ]
        self._joint_dims_210.append(-4)  # -4 is always coil and always a joint dimension

        traj_shape = torch.as_tensor([k.shape[-3:] for k in (traj.kz, traj.ky, traj.kx)])
        self._joint_dims_zyx = []
        for dzyx in [-3, -2, -1]:
            if dzyx not in self._direction_zyx:
                dim210_non_singleton = [d210 for d210 in [-3, -2, -1] if traj_shape[dzyx, d210] > 1]
                if all(all(traj_shape[self._direction_zyx, d] == 1) for d in dim210_non_singleton):
                    self._joint_dims_zyx.append(dzyx)
        self._joint_dims_zyx.append(-4)  # -4 is always coil and always a joint dimension

        self._im_size = im_size
        self._omega = omega.contiguous()

    def _separate_joint_dimensions(
        self, data_ndim: int
    ) -> tuple[Sequence[int], Sequence[int], Sequence[int], Sequence[int]]:
        """Get the separate and joint dimensions for the current data.

        Parameters
        ----------
        data_ndim
            number of dimensions of the data

        Returns
        -------
            separate dimensions along zyx,
            permutation rule for zyx,
            separate dimensions along 210,
            permutation rule for 210.

        """
        # We did most in _init_ and here we only have to check the other dimensions.
        joint_dims_other = []
        for d in range(-data_ndim, -4):
            if abs(d) > len(self._traj_broadcast_shape) or self._traj_broadcast_shape[d] == 1:
                joint_dims_other.append(d)

        sep_dims_zyx = [
            d for d in range(-data_ndim, 0) if d not in [*joint_dims_other, *self._joint_dims_zyx, *self._direction_zyx]
        ]
        permute_zyx = [*sep_dims_zyx, *joint_dims_other, *self._joint_dims_zyx, *self._direction_zyx]
        sep_dims_210 = [
            d for d in range(-data_ndim, 0) if d not in [*joint_dims_other, *self._joint_dims_210, *self._dimension_210]
        ]
        permute_210 = [*sep_dims_210, *joint_dims_other, *self._joint_dims_210, *self._dimension_210]
        return sep_dims_zyx, permute_zyx, sep_dims_210, permute_210

    def __call__(self, x: torch.Tensor) -> tuple[torch.Tensor,]:
        """Apply Non-Uniform Fast Fourier Transform (NUFFT) from image to k-space (Type 2).

        Transforms data from image space to non-uniform k-space locations.

        Parameters
        ----------
        x
            Coil image data, typically with shape `(..., coils, z, y, x)`.

        Returns
        -------
            Coil k-space data at non-uniform locations,
            with shape `(..., coils, k2, k1, k0)`.
        """
        return super().__call__(x)

    def forward(self, x: torch.Tensor) -> tuple[torch.Tensor,]:
        """Apply forward of NonUniformFastFourierOp.

        .. note::
            Prefer calling the instance of the NonUniformFastFourierOp operator as ``operator(x)`` over
            directly calling this method. See this PyTorch `discussion <https://discuss.pytorch.org/t/is-model-forward-x-the-same-as-model-call-x/33460/3>`_.
        """
        if len(self._direction_zyx):
            if x.device.type == 'cpu' and self.oversampling not in (0.0, 1.25, 2.0):
                raise ValueError('Only oversampling 1.25 and 2.0 are supported on CPU')
            elif x.device.type not in ('cuda', 'cpu'):
                raise ValueError('Only CPU and CUDA are supported')
            # We rearrange x into (sep_dims, joint_dims, nufft_directions)
            sep_dims_zyx, permute_zyx, _, permute_210 = self._separate_joint_dimensions(x.ndim)
            unpermute_210 = torch.tensor(permute_210).argsort().tolist()

            x = x.permute(*permute_zyx)
            unflatten_shape = x.shape[: -len(self._direction_zyx)]
            # combine sep_dims
            x = x.flatten(end_dim=len(sep_dims_zyx) - 1) if len(sep_dims_zyx) else x[None, :]
            # combine joint_dims
            x = x.flatten(start_dim=1, end_dim=-len(self._direction_zyx) - 1)
            # cufinufft needs c contiguous, otherwise it warns.
            x = x.contiguous()
            x = torch.vmap(partial(finufft_type2, upsampfac=self.oversampling, modeord=0, isign=-1))(self._omega, x)
            x = x * self.scale
            shape_210 = [self._traj_broadcast_shape[i] for i in self._dimension_210]
            x = x.reshape(*unflatten_shape, *shape_210)
            x = x.permute(*unpermute_210)
        return (x,)

    def adjoint(self, x: torch.Tensor) -> tuple[torch.Tensor,]:
        """Apply adjoint NUFFT from k-space to image space (Type 1).

        Transforms data from non-uniform k-space to a Cartesian image grid.

        Parameters
        ----------
        x
            Coil k-space data at non-uniform locations,
            with shape `(..., coils, k2, k1, k0)`.

        Returns
        -------
            Coil image data on a Cartesian grid,
            with shape `(..., coils, z, y, x)`.
        """
        if len(self._direction_zyx):
            if x.device.type == 'cpu' and self.oversampling not in (0.0, 1.25, 2.0):
                raise ValueError('Only oversampling 1.25 and 2.0 are supported on CPU')
            elif x.device.type not in ('cuda', 'cpu'):
                raise ValueError('Only CPU and CUDA are supported')
            # We rearrange x into (sep_dims, joint_dims, nufft_directions)
            _, permute_zyx, sep_dims_210, permute_210 = self._separate_joint_dimensions(x.ndim)
            unpermute_zyx = torch.tensor(permute_zyx).argsort().tolist()

            x = x.permute(*permute_210)
            unflatten_other_shape = x.shape[: -len(self._dimension_210) - 1]  # -1 for coil
            # combine sep_dims
            x = x.flatten(end_dim=len(sep_dims_210) - 1) if len(sep_dims_210) else x[None, :]
            # combine joint_dims and nufft_dims
            x = x.flatten(start_dim=1, end_dim=-len(self._dimension_210) - 1).flatten(start_dim=2)
            # cufinufft needs c contiguous, otherwise it warns.
            x = x.contiguous()
            x = torch.vmap(
                partial(finufft_type1, upsampfac=self.oversampling, modeord=0, isign=1, output_shape=self._im_size)
            )(self._omega, x)
            x = x * self.scale

            x = x.reshape(*unflatten_other_shape, -1, *x.shape[-len(self._direction_zyx) :])
            x = x.permute(*unpermute_zyx)
        return (x,)

    @property
    def gram(self) -> LinearOperator:
        """Return the gram operator."""
        return NonUniformFastFourierOpGramOp(self)

    def __repr__(self) -> str:
        """Representation method for NUFFT operator."""
        device = self._omega.device if self._omega is not None else 'none'
        zyx = ['z', 'y', 'x']
        k2k1k0 = ['k2', 'k1', 'k0']
        direction_zyx = tuple(zyx[i] for i in self._direction_zyx if i in range(-3, 0))
        dimension_210 = tuple(k2k1k0[i] for i in self._dimension_210 if i in range(-3, 0))
        recon_size_str = ', '.join(f'{dim}={size}' for dim, size in zip(direction_zyx, self._im_size, strict=False))
        direction_zyx_str = ', '.join(direction_zyx)
        dimension_210_str = ', '.join(dimension_210)
        out = (
            f'{type(self).__name__} on device: {device}\n'
            f'Dimension(s) along which NUFFT is applied: ({direction_zyx_str}) / ({dimension_210_str})\n'
            f'Reconstructed image size {recon_size_str}'
        )
        return out


def symmetrize(kernel: torch.Tensor, rank: int) -> torch.Tensor:
    """Enforce hermitian symmetry on the kernel. Returns only half of the kernel."""
    flipped = kernel
    for d in range(-rank, 0):
        flipped = flipped.index_select(d, -1 * torch.arange(flipped.shape[d], device=flipped.device) % flipped.size(d))
    kernel = (kernel + flipped.conj()) / 2
    last_len = kernel.shape[-1]
    return kernel[..., : last_len // 2 + 1]


def gram_nufft_kernel(
    weight: torch.Tensor, trajectory: torch.Tensor, recon_shape: tuple[int] | tuple[int, int] | tuple[int, int, int]
) -> torch.Tensor:
    """Calculate the convolution kernel for the NUFFT gram operator.

    Parameters
    ----------
    weight
        either ones or density compensation weights
    trajectory
        k-space trajectory
    recon_shape
        shape of the reconstructed image

    Returns
    -------
    kernel
        Real valued convolution kernel for `~mrpro.operator.NonUniformFastFourierOpGramOp`, already in Fourier
        space.
    """
    rank = trajectory.shape[-2]
    if weight.dtype.is_complex:
        raise ValueError('Only real weights are supported')
    # Instead of doing one adjoint nufft with double the recon size in all dimensions,
    # we do two adjoint nuffts per dimensions, saving a lot of memory.
    adjnufft = torch.vmap(partial(finufft_type1, modeord=0, isign=1, output_shape=recon_shape))

    kernel = torch.zeros(
        *weight.shape[:2], *[r * 2 for r in recon_shape], dtype=weight.dtype.to_complex(), device=weight.device
    )
    shifts = (torch.tensor(recon_shape) / 2).unsqueeze(-1).to(trajectory)
    for flips in list(product([1, -1], repeat=rank)):
        flipped_trajectory = trajectory * torch.tensor(flips).to(trajectory).unsqueeze(-1)
        kernel_part = adjnufft(flipped_trajectory, torch.polar(weight, (shifts * flipped_trajectory).sum(-2, True)))
        idx: list[slice | EllipsisType] = [...]  # which part of the kernel to is currently being processed
        for dim, flip in zip(range(-rank, 0), flips, strict=True):
            if flip > 0:  # first half in the dimension
                idx.append(slice(0, kernel_part.size(dim)))
            else:  # second half in the dimension
                idx.append(slice(kernel_part.size(dim) + 1, None))
                kernel_part = kernel_part.index_select(
                    dim, torch.arange(kernel_part.size(dim) - 1, 0, -1, device=kernel.device)
                )  # flip
        kernel[tuple(idx)] = kernel_part

    kernel = symmetrize(kernel, rank)
    kernel = torch.fft.hfftn(kernel, dim=list(range(-rank, 0)), norm='backward')
    kernel = torch.fft.fftshift(kernel, dim=list(range(-rank, 0)))
    return kernel


class NonUniformFastFourierOpGramOp(LinearOperator):
    """Gram operator for `NonUniformFastFourierOp`.

    Implements the adjoint of the forward operator of the non-uniform Fast Fourier operator, i.e. the gram operator
    `NUFFT.H@NUFFT`.

    Uses a convolution, implemented as multiplication in Fourier space, to calculate the gram operator
    for the Toeplitz NUFFT operator.

    This should not be used directly, but rather through the `~NonUniformFastFourierOp.gram` method of a
    `NonUniformFastFourierOp` object.
    """

    _kernel: torch.Tensor | None

    def __init__(self, nufft_op: NonUniformFastFourierOp) -> None:
        """Initialize the gram operator.

        Parameters
        ----------
        nufft_op
            The py:class:`NonUniformFastFourierOp` to calculate the gram operator for.

        """
        super().__init__()
        self.nufft_gram: None | LinearOperator = None

        if not nufft_op._dimension_210:
            return

        weight = torch.ones(
            [*nufft_op._traj_broadcast_shape[:-4], 1, *nufft_op._traj_broadcast_shape[-3:]],
        ).to(nufft_op._omega)

        # We rearrange weight into (sep_dims, joint_dims, nufft_dims)
        _, permute_zyx, sep_dims_210, permute_210 = nufft_op._separate_joint_dimensions(weight.ndim)
        unpermute_zyx = torch.tensor(permute_zyx).argsort().tolist()

        weight = weight.permute(*permute_210)
        unflatten_other_shape = weight.shape[: -len(nufft_op._dimension_210) - 1]  # -1 for coil
        # combine sep_dims
        weight = weight.flatten(end_dim=len(sep_dims_210) - 1) if len(sep_dims_210) else weight[None, :]
        # combine joint_dims and nufft_dims
        weight = weight.flatten(start_dim=1, end_dim=-len(nufft_op._dimension_210) - 1).flatten(start_dim=2)

        kernel = gram_nufft_kernel(weight, nufft_op._omega, nufft_op._im_size)
        kernel = kernel.reshape(*unflatten_other_shape, -1, *kernel.shape[-len(nufft_op._direction_zyx) :])
        kernel = kernel.permute(*unpermute_zyx)
        kernel = kernel * (nufft_op.scale) ** 2

        fft = FastFourierOp(
            dim=nufft_op._direction_zyx,
            encoding_matrix=[2 * s for s in nufft_op._im_size],
            recon_matrix=nufft_op._im_size,
        )
        self.nufft_gram = fft.H * kernel @ fft

    def __call__(self, x: torch.Tensor) -> tuple[torch.Tensor,]:
        """Apply the Gram operator of the NonUniformFastFourierOp (NUFFT.H @ NUFFT).

        This operation applies the composition of the adjoint NUFFT operator
        and the forward NUFFT operator. It is implemented via a convolution
        with a precomputed kernel.

        Parameters
        ----------
        x
            Input tensor, typically image-space data with shape `(..., coils, z, y, x)`.

        Returns
        -------
            Output tensor, image-space data after NUFFT.H @ NUFFT has been applied.
        """
        return super().__call__(x)

    def forward(self, x: torch.Tensor) -> tuple[torch.Tensor,]:
        """Apply forward of NonUniformFastFourierOpGramOp.

        .. note::
            Prefer calling the instance of the NonUniformFastFourierOpGramOp operator as ``operator(x)`` over
            directly calling this method. See this PyTorch `discussion <https://discuss.pytorch.org/t/is-model-forward-x-the-same-as-model-call-x/33460/3>`_.
        """
        if self.nufft_gram is not None:
            (x,) = self.nufft_gram(x)

        return (x,)

    def adjoint(self, x: torch.Tensor) -> tuple[torch.Tensor,]:
        """Apply the adjoint of the Gram operator.

        Since the Gram operator (NUFFT.H @ NUFFT) is self-adjoint,
        this method calls the forward operation.

        Parameters
        ----------
        x
            Input tensor, typically image-space data with shape `(..., coils, z, y, x)`.

        Returns
        -------
            Output tensor, same shape as the input.
        """
        return self.forward(x)

    @property
    def H(self) -> Self:  # noqa: N802
        """Adjoint operator of the gram operator."""
        return self



"""Structural Similarity Index (SSIM) functional."""

from typing import Literal, cast

import torch

from mrpro.operators.Operator import Operator
from mrpro.utils.sliding_window import sliding_window


def ssim3d(
    target: torch.Tensor,
    prediction: torch.Tensor,
    *,
    data_range: torch.Tensor | None = None,
    weight: torch.Tensor | None = None,
    k1: float = 0.01,
    k2: float = 0.03,
    window_size: int = 7,
    reduction: Literal['full', 'volume', 'none'] = 'full',
) -> torch.Tensor:
    """Compute SSIM between two 3D volumes.

    For complex inputs, the SSIM is calculated separately for the real and imaginary parts and the results are averaged.
    For more details, see `SSIM`.

    Parameters
    ----------
    target
        Ground truth tensor, shape `(... z, y, x)` or broadcastable with prediction.
    prediction
        Predicted tensor, same shape as target.
    data_range
        Value range if the data. If `None`, the max-to-min per volume of the target will be used.
    weight
        Weight (or mask) tensor, same shape as target.
        Only windows with all weight values > 0 (or `True`) are considered.
        Windows will be weighted by the average value of the weight in the window.
        If `None`, all windows are considered without weighting.
    k1
        Constant for SSIM computation. Commonly 0.01.
    k2
        Constant for SSIM computation. Commonly 0.03.
    window_size
        Size of the rectangular sliding window.
        If any of the last 3 dimensions of the target is of size 1, the window size in this dimension will
        also be set to 1.

    reduction
        If `full`, return the weighted mean SSIM over all windows, i.e. a scalar value.
        If `volume`, return one value for each volume, i.e, an ``target.ndim - 3`` dimensional tensor.
        If `none`, return the unpadded SSIM map of shape
        ``(... z - window_size + 1 ,  y - window_size + 1 , x - window_size + 1)``.

    Returns
    -------
        mean SSIM or SSIM map (see `reduction`)
    """
    if target.is_complex() or prediction.is_complex():
        real_ssim = ssim3d(
            target.real,
            prediction.real,
            weight=weight,
            k1=k1,
            k2=k2,
            window_size=window_size,
            data_range=data_range,
            reduction=reduction,
        )
        imag_ssim = ssim3d(
            target.imag if target.is_complex() else torch.zeros_like(target),
            prediction.imag if prediction.is_complex() else torch.zeros_like(prediction),
            weight=weight,
            k1=k1,
            k2=k2,
            window_size=window_size,
            data_range=data_range,
            reduction=reduction,
        )
        return (real_ssim + imag_ssim) / 2
    if target.ndim < 3:
        raise ValueError('Input must be at least 3D (z, y, x)')
    window = tuple(window_size if s > 1 else 1 for s in target.shape[-3:])  # To support 1D and 2D uses
    if weight is not None:
        if (weight < 0).any():
            raise ValueError('Mask contains negative values')
        target, prediction, weight = cast(
            tuple[torch.Tensor, torch.Tensor, torch.Tensor], torch.broadcast_tensors(target, prediction, weight)
        )
        weight = sliding_window(weight, window_shape=window, dim=(-3, -2, -1))
        # Set weights to 0 for windows that are not fully inside the mask
        weight = weight * ~torch.isclose(weight, torch.tensor(0, dtype=weight.dtype)).any((-3, -2, -1), keepdim=True)
        weight = weight.mean((-1, -2, -3), dtype=torch.float32).moveaxis((0, 1, 2), (-3, -2, -1))
        weight /= weight.sum(dim=(-3, -2, -1), keepdim=True)  # Normalization for mean

    else:
        target, prediction = cast(tuple[torch.Tensor, torch.Tensor], torch.broadcast_tensors(target, prediction))

    target_window = sliding_window(target, window_shape=window, dim=(-3, -2, -1)).movedim((0, 1, 2), (-6, -5, -4))

    if data_range is None:
        if weight is None:
            target_max = target_window.amax((-3, -2, -1)).amax((-3, -2, -1), keepdim=True)
            target_min = target_window.amin((-3, -2, -1)).amin((-3, -2, -1), keepdim=True)
        else:
            target_max = torch.where(weight > 0, target_window.amax((-3, -2, -1)), -torch.inf).amax(
                (-3, -2, -1), keepdim=True
            )
            target_min = torch.where(weight > 0, target_window.amin((-3, -2, -1)), torch.inf).amin(
                (-3, -2, -1), keepdim=True
            )
        data_range_ = target_max - target_min
    else:
        data_range_ = data_range

    x_window = sliding_window(prediction, window_shape=window, dim=(-3, -2, -1)).movedim((0, 1, 2), (-6, -5, -4))

    mean_tgt = target_window.mean(dim=(-3, -2, -1))
    mean_img = x_window.mean(dim=(-3, -2, -1))

    mean_tgt_tgt = (target_window**2).mean(dim=(-3, -2, -1))
    mean_img_img = (x_window**2).mean(dim=(-3, -2, -1))
    mean_tgt_img = (target_window * x_window).mean(dim=(-3, -2, -1))

    n = x_window.shape[-3:].numel()
    cov_norm = n / (n - 1)
    cov_tgt = cov_norm * (mean_tgt_tgt - mean_tgt * mean_tgt)
    cov_img = cov_norm * (mean_img_img - mean_img * mean_img)
    cov_tgt_img = cov_norm * (mean_tgt_img - mean_tgt * mean_img)

    c1 = (k1 * data_range_) ** 2
    c2 = (k2 * data_range_) ** 2
    a1 = 2 * mean_tgt * mean_img + c1
    a2 = 2 * cov_tgt_img + c2
    b1 = mean_tgt**2 + mean_img**2 + c1
    b2 = cov_tgt + cov_img + c2

    ssim_map = (a1 * a2) / (b1 * b2)

    if reduction == 'full':
        if weight is not None:
            return (ssim_map * weight).sum((-3, -2, -1)).mean()
        else:
            return ssim_map.mean()
    elif reduction == 'volume':
        if weight is not None:
            return (ssim_map * weight).sum(dim=(-3, -2, -1))
        else:
            return ssim_map.mean(dim=(-3, -2, -1))
    elif reduction == 'none':
        return ssim_map


class SSIM(Operator[torch.Tensor, tuple[torch.Tensor]]):
    """(masked) SSIM functional."""

    def __init__(
        self,
        target: torch.Tensor,
        weight: torch.Tensor | None = None,
        *,
        data_range: torch.Tensor | None = None,
        window_size: int = 7,
        k1: float = 0.01,
        k2: float = 0.03,
        reduction: Literal['full', 'volume', 'none'] = 'full',
    ) -> None:
        """Initialize Volume SSIM.

        The Structural Similarity Index Measure [SSIM]_ is used to measure the similarity between two images.
        It considers luminance, contrast and structure differences between the images.
        SSIM values range from -1 to 1, where 1 indicates perfect structural similarity.
        Two random images have an SSIM of 0.

        Calculates the SSIM using a rectangular sliding window. If a boolean mask is used as weight, only the windows
        that are fully inside the mask are considered. This can be used to ignore the background of the volumes.
        Calculates the SSIM for a volume, i.e, 3D patches of the last three dimensions of the input are used.
        To apply it to 2D data, add a singleton dimension.
        For complex inputs, the SSIM is calculated separately for the real and imaginary parts and the results
        are averaged.

        For stability, it is advised to provide the data range. Otherwise it is estimates per volume from the target.

        References
        ----------
        .. [SSIM] Wang, Z., Bovik, A. C., Sheikh, H. R., & Simoncelli, E. P. (2004).
               Image quality assessment: from error visibility to structural similarity.
               IEEE TMI, 13(4), 600-612. https://doi.org/10.1109/TIP.2003.819861


        Parameters
        ----------
        target
            Target volume. At least three dimensional.
        weight
            Either a boolean mask. Only windows with all values `True` will be considered.
            Or a weight tensor of the same shape as the target. Each window will be weighted by
            the average value of the weight in the window. Only windows with all weight values > 0 will be considered.
            Or `None`, meaning all windows are used.
        data_range
            Value range if the data. If None, the max-to-min per volume of the target will be used.
        window_size
            Size of the windows used in SSIM. Usually ``7`` or ``11``.
            If any of the last 3 dimensions of the target is of size 1, the window size in this dimension will
            also be set to 1.
        k1
            Constant. Usually ``0.01`` and rarely changed.
        k2
            Constant. Usually ``0.03`` and rarely changed.
        reduction
            If ``full``, return the weighted mean SSIM over all windows, i.e. a scalar value.
            If ``volume``, return one value for each volume, i.e, an ``target.ndim - 3`` dimensional tensor.
            If ``none``, return the unpadded SSIM map.
        """
        super().__init__()
        self.target = target
        self.weight = weight
        self.k1 = k1
        self.k2 = k2
        self.window_size = window_size
        self.data_range = data_range
        self.reduction = reduction

    def __call__(self, x: torch.Tensor) -> tuple[torch.Tensor]:
        """Calculate the Structural Similarity Index (SSIM) between the input and a target.

        This method computes the SSIM between the provided input tensor `x` and
        the `target` tensor defined during initialization.

        Parameters
        ----------
        x
            Input tensor, expected to be comparable to `target`.

        Returns
        -------
            The SSIM value(s), the shape of which depends on the `reduction` parameter.
        """
        return super().__call__(x)

    def forward(self, x: torch.Tensor) -> tuple[torch.Tensor]:
        """Apply forward of SSIM.

        .. note::
            Prefer calling the instance of the SSIM as ``operator(x)`` over directly calling this method.
            See this PyTorch `discussion <https://discuss.pytorch.org/t/is-model-forward-x-the-same-as-model-call-x/33460/3>`_.
        """
        ssim = ssim3d(
            self.target.real,
            x.real,
            weight=self.weight,
            k1=self.k1,
            k2=self.k2,
            window_size=self.window_size,
            data_range=self.data_range,
            reduction=self.reduction,
        )

        return (ssim,)



"""All acquisition enums."""

from enum import Enum, Flag, IntFlag, auto


class AcqFlags(Flag):
    """Acquisition flags.

    NOTE: values in enum ISMRMRD_AcquisitionFlags start at 1 and not 0, but
    1 << (val-1) is used in 'ismrmrd_is_flag_set' function to calc bitmask value [ISMb]_.

    References
    ----------
    .. [ISMb] ISMRMRD https://github.com/ismrmrd/ismrmrd/blob/master/include/ismrmrd/ismrmrd.h
    """

    ACQ_NO_FLAG = 0
    ACQ_FIRST_IN_ENCODE_STEP1 = auto()
    ACQ_LAST_IN_ENCODE_STEP1 = auto()
    ACQ_FIRST_IN_ENCODE_STEP2 = auto()
    ACQ_LAST_IN_ENCODE_STEP2 = auto()
    ACQ_FIRST_IN_AVERAGE = auto()
    ACQ_LAST_IN_AVERAGE = auto()
    ACQ_FIRST_IN_SLICE = auto()
    ACQ_LAST_IN_SLICE = auto()
    ACQ_FIRST_IN_CONTRAST = auto()
    ACQ_LAST_IN_CONTRAST = auto()
    ACQ_FIRST_IN_PHASE = auto()
    ACQ_LAST_IN_PHASE = auto()
    ACQ_FIRST_IN_REPETITION = auto()
    ACQ_LAST_IN_REPETITION = auto()
    ACQ_FIRST_IN_SET = auto()
    ACQ_LAST_IN_SET = auto()
    ACQ_FIRST_IN_SEGMENT = auto()
    ACQ_LAST_IN_SEGMENT = auto()
    ACQ_IS_NOISE_MEASUREMENT = auto()
    ACQ_IS_PARALLEL_CALIBRATION = auto()
    ACQ_IS_PARALLEL_CALIBRATION_AND_IMAGING = auto()
    ACQ_IS_REVERSE = auto()
    ACQ_IS_NAVIGATION_DATA = auto()
    ACQ_IS_PHASECORR_DATA = auto()
    ACQ_LAST_IN_MEASUREMENT = auto()
    ACQ_IS_HPFEEDBACK_DATA = auto()
    ACQ_IS_DUMMYSCAN_DATA = auto()
    ACQ_IS_RTFEEDBACK_DATA = auto()
    ACQ_IS_SURFACECOILCORRECTIONSCAN_DATA = auto()
    ACQ_IS_PHASE_STABILIZATION_REFERENCE = auto()
    ACQ_IS_PHASE_STABILIZATION = 30 << 1
    ACQ_COMPRESSION1 = 52 << 1  # 53 (on purpose!)
    ACQ_COMPRESSION2 = auto()
    ACQ_COMPRESSION3 = auto()
    ACQ_COMPRESSION4 = auto()
    ACQ_USER1 = auto()
    ACQ_USER2 = auto()
    ACQ_USER3 = auto()
    ACQ_USER4 = auto()
    ACQ_USER5 = auto()
    ACQ_USER6 = auto()
    ACQ_USER7 = auto()
    ACQ_USER8 = auto()


class InterleavingDimension(Enum):
    """Interleaving dimension."""

    PHASE = 'phase'
    REPETITION = 'repetition'
    CONTRAST = 'contrast'
    AVERAGE = 'average'
    OTHER = 'other'


class MultibandCalibration(Enum):
    """Multiband calibration."""

    SEPARABLE2_D = 'separable2D'
    FULL3_D = 'full3D'
    OTHER = 'other'


class PatientPosition(Enum):
    """Patient position."""

    HFP = 'HFP'
    HFS = 'HFS'
    HFDR = 'HFDR'
    HFDL = 'HFDL'
    FFP = 'FFP'
    FFS = 'FFS'
    FFDR = 'FFDR'
    FFDL = 'FFDL'
    OTHER = 'OTHER'


class TrajectoryType(Enum):
    """Trajectory type."""

    CARTESIAN = 'cartesian'
    EPI = 'epi'
    RADIAL = 'radial'
    GOLDENANGLE = 'goldenangle'
    SPIRAL = 'spiral'
    OTHER = 'other'


class WaveformInformation(Enum):
    """Waveform information."""

    ECG = 'ecg'
    PULSE = 'pulse'
    RESPIRATORY = 'respiratory'
    TRIGGER = 'trigger'
    GRADIENTWAVEFORM = 'gradientwaveform'
    OTHER = 'other'


class CalibrationMode(Enum):
    """Calibration mode."""

    EMBEDDED = 'embedded'
    INTERLEAVED = 'interleaved'
    SEPARATE = 'separate'
    EXTERNAL = 'external'
    OTHER = 'other'


class TrajType(IntFlag):
    """Special Properties of the Trajectory."""

    SINGLEVALUE = auto()
    ONGRID = auto()



"""reduce_repeat utility function."""

from collections.abc import Sequence

import torch


def reduce_repeat(tensor: torch.Tensor, tol: float = 1e-6, dim: Sequence[int] | None = None) -> torch.Tensor:
    """Replace dimensions with all equal values with singletons.

    Parameters
    ----------
    tensor
        Input tensor
    tol
        tolerance.
    dim
        dimensions to try to reduce to singletons. `None` means all.
    """
    if tensor.is_complex():
        real = reduce_repeat(tensor.real, tol, dim)
        imag = reduce_repeat(tensor.imag, tol, dim)
        return real + 1j * imag

    if dim is not None:
        dim = [d % tensor.ndim for d in dim]

    def can_be_singleton(d: int) -> bool:
        if dim is not None and d not in dim:
            # not in the list of dimensions to reduce
            return False
        if tensor.stride(d) == 0:
            # broadcasted dimension
            return True
        if tol >= 0 and tensor.dtype == torch.bool:
            # either all true or all false
            return bool((tensor.all(dim=d) | ~tensor.any(dim=d)).all())

        # If the distance between min and max is smaller than the tolerance, all values are the same.
        return bool(torch.all((tensor.amax(dim=d) - tensor.amin(dim=d)) <= tol).item())

    take_first = slice(0, 1)
    take_all = slice(None)
    index = tuple(take_first if can_be_singleton(dim) else take_all for dim in range(tensor.ndim))
    return tensor[index]
